# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Action base class and some of his extentions.
An Action has a name, a list of Parameter, a list of preconditions
and a list of effects.
"""


import unified_planning as up
from unified_planning.environment import get_env, Environment
from unified_planning.exceptions import (
    UPTypeError,
    UPUnboundedVariablesError,
    UPProblemDefinitionError,
    UPConflictingEffectsException,
)
from fractions import Fraction
from typing import Dict, List, Set, Tuple, Union, Optional
from collections import OrderedDict


class Action:
    """This is the action interface."""

    def __init__(
        self,
        _name: str,
        _parameters: "OrderedDict[str, up.model.types.Type]" = None,
        _env: Environment = None,
        **kwargs: "up.model.types.Type",
    ):
        self._env = get_env(_env)
        self._name = _name
        self._parameters: "OrderedDict[str, up.model.parameter.Parameter]" = (
            OrderedDict()
        )
        if _parameters is not None:
            assert len(kwargs) == 0
            for n, t in _parameters.items():
                assert self._env.type_manager.has_type(
                    t
                ), "type of parameter does not belong to the same environment of the action"
                self._parameters[n] = up.model.parameter.Parameter(n, t, self._env)
        else:
            for n, t in kwargs.items():
                assert self._env.type_manager.has_type(
                    t
                ), "type of parameter does not belong to the same environment of the action"
                self._parameters[n] = up.model.parameter.Parameter(n, t, self._env)

    def __eq__(self, oth: object) -> bool:
        raise NotImplementedError

    def __hash__(self) -> int:
        raise NotImplementedError

    def clone(self):
        raise NotImplementedError

    @property
    def env(self) -> Environment:
        """Returns this action environment."""
        return self._env

    @property
    def name(self) -> str:
        """Returns the action name."""
        return self._name

    @name.setter
    def name(self, new_name: str):
        """Sets the parameter name."""
        self._name = new_name

    @property
    def parameters(self) -> List["up.model.parameter.Parameter"]:
        """Returns the list of the action parameters."""
        return list(self._parameters.values())

    def parameter(self, name: str) -> "up.model.parameter.Parameter":
        """Returns the parameter of the action with the given name."""
        return self._parameters[name]

    def is_conditional(self) -> bool:
        """Returns True if the action has conditional effects."""
        raise NotImplementedError


class InstantaneousAction(Action):
    """Represents an instantaneous action."""

    def __init__(
        self,
        _name: str,
        _parameters: "OrderedDict[str, up.model.types.Type]" = None,
        _env: Environment = None,
        **kwargs: "up.model.types.Type",
    ):
        Action.__init__(self, _name, _parameters, _env, **kwargs)
        self._preconditions: List[up.model.fnode.FNode] = []
        self._effects: List[up.model.effect.Effect] = []
        self._simulated_effect: Optional[up.model.effect.SimulatedEffect] = None
        # fluent assigned is the set of the fluent that have an unconditional assignment
        self._fluents_assigned: Set["up.model.FNode"] = set()
        # fluent_inc_dec is the set of the fluents that have an unconditional increase or decrease
        self._fluents_inc_dec: Set["up.model.FNode"] = set()

    def __repr__(self) -> str:
        s = []
        s.append(f"action {self.name}")
        first = True
        for p in self.parameters:
            if first:
                s.append("(")
                first = False
            else:
                s.append(", ")
            s.append(str(p))
        if not first:
            s.append(")")
        s.append(" {\n")
        s.append("    preconditions = [\n")
        for c in self.preconditions:
            s.append(f"      {str(c)}\n")
        s.append("    ]\n")
        s.append("    effects = [\n")
        for e in self.effects:
            s.append(f"      {str(e)}\n")
        s.append("    ]\n")
        s.append(f"    simulated effect = {self._simulated_effect}\n")
        s.append("  }")
        return "".join(s)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, InstantaneousAction):
            cond = (
                self._env == oth._env
                and self._name == oth._name
                and self._parameters == oth._parameters
            )
            return (
                cond
                and set(self._preconditions) == set(oth._preconditions)
                and set(self._effects) == set(oth._effects)
                and self._simulated_effect == oth._simulated_effect
            )
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._name)
        for ap in self._parameters.items():
            res += hash(ap)
        for p in self._preconditions:
            res += hash(p)
        for e in self._effects:
            res += hash(e)
        res += hash(self._simulated_effect)
        return res

    def clone(self):
        new_params = {}
        for param_name, param in self._parameters.items():
            new_params[param_name] = param.type
        new_instantaneous_action = InstantaneousAction(
            self._name, new_params, self._env
        )
        new_instantaneous_action._preconditions = self._preconditions[:]
        new_instantaneous_action._effects = [e.clone() for e in self._effects]
        new_instantaneous_action._fluents_assigned = self._fluents_assigned.copy()
        new_instantaneous_action._fluents_inc_dec = self._fluents_inc_dec.copy()
        new_instantaneous_action._simulated_effect = self._simulated_effect
        return new_instantaneous_action

    @property
    def preconditions(self) -> List["up.model.fnode.FNode"]:
        """Returns the list of the action preconditions."""
        return self._preconditions

    def clear_preconditions(self):
        """Removes all action preconditions"""
        self._preconditions = []

    @property
    def effects(self) -> List["up.model.effect.Effect"]:
        """Returns the list of the action effects."""
        return self._effects

    def clear_effects(self):
        """Removes all effects."""
        self._effects = []
        self._fluents_assigned = set()
        self._fluents_inc_dec = set()

    @property
    def conditional_effects(self) -> List["up.model.effect.Effect"]:
        """Returns the list of the action conditional effects.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        return [e for e in self._effects if e.is_conditional()]

    def is_conditional(self) -> bool:
        """Returns True if the action has conditional effects."""
        return any(e.is_conditional() for e in self._effects)

    @property
    def unconditional_effects(self) -> List["up.model.effect.Effect"]:
        """Returns the list of the action unconditional effects.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        return [e for e in self._effects if not e.is_conditional()]

    def add_precondition(
        self,
        precondition: Union[
            "up.model.fnode.FNode",
            "up.model.fluent.Fluent",
            "up.model.parameter.Parameter",
            bool,
        ],
    ):
        """Adds the given action precondition."""
        (precondition_exp,) = self._env.expression_manager.auto_promote(precondition)
        assert self._env.type_checker.get_type(precondition_exp).is_bool_type()
        if precondition_exp == self._env.expression_manager.TRUE():
            return
        free_vars = self._env.free_vars_oracle.get_free_variables(precondition_exp)
        if len(free_vars) != 0:
            raise UPUnboundedVariablesError(
                f"The precondition {str(precondition_exp)} has unbounded variables:\n{str(free_vars)}"
            )
        if precondition_exp not in self._preconditions:
            self._preconditions.append(precondition_exp)

    def add_effect(
        self,
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given action effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("InstantaneousAction effect has not compatible types!")
        self._add_effect_instance(
            up.model.effect.Effect(fluent_exp, value_exp, condition_exp)
        )

    def add_increase_effect(
        self,
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given action increase effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("InstantaneousAction effect has not compatible types!")
        self._add_effect_instance(
            up.model.effect.Effect(
                fluent_exp,
                value_exp,
                condition_exp,
                kind=up.model.effect.EffectKind.INCREASE,
            )
        )

    def add_decrease_effect(
        self,
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given action decrease effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("InstantaneousAction effect has not compatible types!")
        self._add_effect_instance(
            up.model.effect.Effect(
                fluent_exp,
                value_exp,
                condition_exp,
                kind=up.model.effect.EffectKind.DECREASE,
            )
        )

    def _add_effect_instance(self, effect: "up.model.effect.Effect"):
        assert (
            effect.environment == self._env
        ), "effect does not have the same environment of the action"
        if not effect.is_conditional():
            if effect.is_assignment():
                if (
                    effect.fluent in self._fluents_assigned
                    or effect.fluent in self._fluents_inc_dec
                ):
                    raise UPConflictingEffectsException(
                        f"The effect {effect} is in conflict with the effects already in the action."
                    )
                self._fluents_assigned.add(effect.fluent)
            elif effect.is_increase() or effect.is_decrease():
                if effect.fluent in self._fluents_assigned:
                    raise UPConflictingEffectsException(
                        f"The effect {effect} is in conflict with the effects already in the action."
                    )
                self._fluents_inc_dec.add(effect.fluent)
            else:
                raise NotImplementedError
        if (
            self._simulated_effect is not None
            and effect.fluent in self._simulated_effect.fluents
        ):
            raise UPConflictingEffectsException(
                f"The effect {effect} is in conflict with the simulated effects already in the action."
            )
        self._effects.append(effect)

    @property
    def simulated_effect(self) -> Optional["up.model.effect.SimulatedEffect"]:
        """Returns the action simulated effect."""
        return self._simulated_effect

    def set_simulated_effect(self, simulated_effect: "up.model.effect.SimulatedEffect"):
        """Sets the given simulated effect."""
        for f in simulated_effect.fluents:
            if f in self._fluents_assigned or f in self._fluents_inc_dec:
                raise UPConflictingEffectsException(
                    f"The simulated effect {simulated_effect} is in conflict with the effects already in the action."
                )
        self._simulated_effect = simulated_effect

    def _set_preconditions(self, preconditions: List["up.model.fnode.FNode"]):
        self._preconditions = preconditions


class DurativeAction(Action):
    """Represents a durative action."""

    def __init__(
        self,
        _name: str,
        _parameters: "OrderedDict[str, up.model.types.Type]" = None,
        _env: Environment = None,
        **kwargs: "up.model.types.Type",
    ):
        Action.__init__(self, _name, _parameters, _env, **kwargs)
        self._duration: "up.model.timing.DurationInterval" = (
            up.model.timing.FixedDuration(self._env.expression_manager.Int(0))
        )
        self._conditions: Dict[
            "up.model.timing.TimeInterval", List["up.model.fnode.FNode"]
        ] = {}
        self._effects: Dict[
            "up.model.timing.Timing", List["up.model.effect.Effect"]
        ] = {}
        self._simulated_effects: Dict[
            "up.model.timing.Timing", "up.model.effect.SimulatedEffect"
        ] = {}
        self._fluents_assigned: Dict[
            "up.model.timing.Timing", Set["up.model.FNode"]
        ] = {}
        self._fluents_inc_dec: Dict[
            "up.model.timing.Timing", Set["up.model.FNode"]
        ] = {}

    def __repr__(self) -> str:
        s = []
        s.append(f"durative action {self.name}")
        first = True
        for p in self.parameters:
            if first:
                s.append("(")
                first = False
            else:
                s.append(", ")
            s.append(str(p))
        if not first:
            s.append(")")
        s.append(" {\n")
        s.append(f"    duration = {str(self._duration)}\n")
        s.append("    conditions = [\n")
        for i, cl in self.conditions.items():
            s.append(f"      {str(i)}:\n")
            for c in cl:
                s.append(f"        {str(c)}\n")
        s.append("    ]\n")
        s.append("    effects = [\n")
        for t, el in self.effects.items():
            s.append(f"      {str(t)}:\n")
            for e in el:
                s.append(f"        {str(e)}:\n")
        s.append("    ]\n")
        s.append("    simulated effects = [\n")
        for t, se in self.simulated_effects.items():
            s.append(f"      {str(t)}: {se}\n")
        s.append("    ]\n")
        s.append("  }")
        return "".join(s)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, DurativeAction):
            if (
                self._env != oth._env
                or self._name != oth._name
                or self._parameters != oth._parameters
                or self._duration != oth._duration
            ):
                return False
            if len(self._conditions) != len(oth._conditions):
                return False
            for i, cl in self._conditions.items():
                oth_cl = oth._conditions.get(i, None)
                if oth_cl is None:
                    return False
                elif set(cl) != set(oth_cl):
                    return False
            if len(self._effects) != len(oth._effects):
                return False
            for t, el in self._effects.items():
                oth_el = oth._effects.get(t, None)
                if oth_el is None:
                    return False
                elif set(el) != set(oth_el):
                    return False
            for t, se in self._simulated_effects.items():
                oth_se = oth._simulated_effects.get(t, None)
                if oth_se is None:
                    return False
                elif se != oth_se:
                    return False
            return True
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._name) + hash(self._duration)
        for ap in self._parameters.items():
            res += hash(ap)
        for i, cl in self._conditions.items():
            res += hash(i)
            for c in cl:
                res += hash(c)
        for t, el in self._effects.items():
            res += hash(t)
            for e in el:
                res += hash(e)
        for t, se in self._simulated_effects.items():
            res += hash(t) + hash(se)
        return res

    def clone(self):
        new_params = {
            param_name: param.type for param_name, param in self._parameters.items()
        }
        new_durative_action = DurativeAction(self._name, new_params, self._env)
        new_durative_action._duration = self._duration
        new_durative_action._conditions = {
            t: cl[:] for t, cl in self._conditions.items()
        }
        new_durative_action._effects = {
            t: [e.clone() for e in el] for t, el in self._effects.items()
        }
        new_durative_action._simulated_effects = {
            t: se for t, se in self._simulated_effects.items()
        }
        new_durative_action._fluents_assigned = {
            t: fs.copy() for t, fs in self._fluents_assigned.items()
        }
        new_durative_action._fluents_inc_dec = {
            t: fs.copy() for t, fs in self._fluents_inc_dec.items()
        }
        return new_durative_action

    @property
    def duration(self) -> "up.model.timing.DurationInterval":
        """Returns the action duration interval."""
        return self._duration

    @property
    def conditions(
        self,
    ) -> Dict["up.model.timing.TimeInterval", List["up.model.fnode.FNode"]]:
        """Returns the action conditions."""
        return self._conditions

    def clear_conditions(self):
        """Removes all conditions."""
        self._conditions = {}

    @property
    def effects(self) -> Dict["up.model.timing.Timing", List["up.model.effect.Effect"]]:
        """Returns the action effects."""
        return self._effects

    def clear_effects(self):
        """Removes all effects."""
        self._effects = {}
        self._fluents_assigned = {}
        self._fluents_inc_dec = {}

    @property
    def conditional_effects(
        self,
    ) -> Dict["up.model.timing.Timing", List["up.model.effect.Effect"]]:
        """Return the action conditional effects.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        retval: Dict[up.model.timing.Timing, List[up.model.effect.Effect]] = {}
        for timing, effect_list in self._effects.items():
            cond_effect_list = [e for e in effect_list if e.is_conditional()]
            if len(cond_effect_list) > 0:
                retval[timing] = cond_effect_list
        return retval

    @property
    def unconditional_effects(
        self,
    ) -> Dict["up.model.timing.Timing", List["up.model.effect.Effect"]]:
        """Return the action unconditional effects.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        retval: Dict[up.model.timing.Timing, List[up.model.effect.Effect]] = {}
        for timing, effect_list in self._effects.items():
            uncond_effect_list = [e for e in effect_list if not e.is_conditional()]
            if len(uncond_effect_list) > 0:
                retval[timing] = uncond_effect_list
        return retval

    def is_conditional(self) -> bool:
        """Returns True if the action has conditional effects."""
        return any(e.is_conditional() for l in self._effects.values() for e in l)

    def set_duration_constraint(self, duration: "up.model.timing.DurationInterval"):
        """Sets the duration interval."""
        lower, upper = duration.lower, duration.upper
        tlower = self._env.type_checker.get_type(lower)
        tupper = self._env.type_checker.get_type(upper)
        assert tlower.is_int_type() or tlower.is_real_type()
        assert tupper.is_int_type() or tupper.is_real_type()
        if (
            lower.is_constant()
            and upper.is_constant()
            and (
                upper.constant_value() < lower.constant_value()
                or (
                    upper.constant_value() == lower.constant_value()
                    and (duration.is_left_open() or duration.is_right_open())
                )
            )
        ):
            raise UPProblemDefinitionError(
                f"{duration} is an empty interval duration of action: {self.name}."
            )
        self._duration = duration

    def set_fixed_duration(self, value: Union["up.model.fnode.FNode", int, Fraction]):
        (value_exp,) = self._env.expression_manager.auto_promote(value)
        self.set_duration_constraint(up.model.timing.FixedDuration(value_exp))

    def set_closed_duration_interval(
        self,
        lower: Union["up.model.fnode.FNode", int, Fraction],
        upper: Union["up.model.fnode.FNode", int, Fraction],
    ):
        lower_exp, upper_exp = self._env.expression_manager.auto_promote(lower, upper)
        self.set_duration_constraint(
            up.model.timing.ClosedDurationInterval(lower_exp, upper_exp)
        )

    def set_open_duration_interval(
        self,
        lower: Union["up.model.fnode.FNode", int, Fraction],
        upper: Union["up.model.fnode.FNode", int, Fraction],
    ):
        lower_exp, upper_exp = self._env.expression_manager.auto_promote(lower, upper)
        self.set_duration_constraint(
            up.model.timing.OpenDurationInterval(lower_exp, upper_exp)
        )

    def set_left_open_duration_interval(
        self,
        lower: Union["up.model.fnode.FNode", int, Fraction],
        upper: Union["up.model.fnode.FNode", int, Fraction],
    ):
        lower_exp, upper_exp = self._env.expression_manager.auto_promote(lower, upper)
        self.set_duration_constraint(
            up.model.timing.LeftOpenDurationInterval(lower_exp, upper_exp)
        )

    def set_right_open_duration_interval(
        self,
        lower: Union["up.model.fnode.FNode", int, Fraction],
        upper: Union["up.model.fnode.FNode", int, Fraction],
    ):
        lower_exp, upper_exp = self._env.expression_manager.auto_promote(lower, upper)
        self.set_duration_constraint(
            up.model.timing.RightOpenDurationInterval(lower_exp, upper_exp)
        )

    def add_condition(
        self,
        interval: Union["up.model.timing.Timing", "up.model.timing.TimeInterval"],
        condition: Union[
            "up.model.fnode.FNode",
            "up.model.fluent.Fluent",
            "up.model.parameter.Parameter",
            bool,
        ],
    ):
        """Adds the given condition."""
        if isinstance(interval, up.model.Timing):
            interval = up.model.TimePointInterval(interval)
        (condition_exp,) = self._env.expression_manager.auto_promote(condition)
        assert self._env.type_checker.get_type(condition_exp).is_bool_type()
        if interval in self._conditions:
            if condition_exp not in self._conditions[interval]:
                self._conditions[interval].append(condition_exp)
        else:
            self._conditions[interval] = [condition_exp]

    def _set_conditions(
        self,
        interval: "up.model.timing.TimeInterval",
        conditions: List["up.model.fnode.FNode"],
    ):
        self._conditions[interval] = conditions

    def add_effect(
        self,
        timing: "up.model.timing.Timing",
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given action effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("InstantaneousAction effect has not compatible types!")
        self._add_effect_instance(
            timing, up.model.effect.Effect(fluent_exp, value_exp, condition_exp)
        )

    def add_increase_effect(
        self,
        timing: "up.model.timing.Timing",
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given action increase effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("InstantaneousAction effect has not compatible types!")
        self._add_effect_instance(
            timing,
            up.model.effect.Effect(
                fluent_exp,
                value_exp,
                condition_exp,
                kind=up.model.effect.EffectKind.INCREASE,
            ),
        )

    def add_decrease_effect(
        self,
        timing: "up.model.timing.Timing",
        fluent: Union["up.model.fnode.FNode", "up.model.fluent.Fluent"],
        value: "up.model.expression.Expression",
        condition: "up.model.expression.BoolExpression" = True,
    ):
        """Adds the given action decrease effect."""
        (
            fluent_exp,
            value_exp,
            condition_exp,
        ) = self._env.expression_manager.auto_promote(fluent, value, condition)
        assert fluent_exp.is_fluent_exp()
        if not self._env.type_checker.get_type(condition_exp).is_bool_type():
            raise UPTypeError("Effect condition is not a Boolean condition!")
        if not fluent_exp.type.is_compatible(value_exp.type):
            raise UPTypeError("InstantaneousAction effect has not compatible types!")
        self._add_effect_instance(
            timing,
            up.model.effect.Effect(
                fluent_exp,
                value_exp,
                condition_exp,
                kind=up.model.effect.EffectKind.DECREASE,
            ),
        )

    def _add_effect_instance(
        self, timing: "up.model.timing.Timing", effect: "up.model.effect.Effect"
    ):
        assert (
            self._env == effect.environment
        ), "effect does not have the same environment of the action"
        fluents_assigned = self._fluents_assigned.setdefault(timing, set())
        fluents_inc_dec = self._fluents_inc_dec.setdefault(timing, set())
        simulated_effect = self._simulated_effects.get(timing, None)
        if not effect.is_conditional():
            if effect.is_assignment():
                if (
                    effect.fluent in fluents_assigned
                    or effect.fluent in fluents_inc_dec
                ):
                    raise UPConflictingEffectsException(
                        f"The effect {effect} at timing {timing} is in conflict with the effects already in the action."
                    )
                fluents_assigned.add(effect.fluent)
            elif effect.is_increase() or effect.is_decrease():
                if effect.fluent in fluents_assigned:
                    raise UPConflictingEffectsException(
                        f"The effect {effect} at timing {timing} is in conflict with the effects already in the action."
                    )
                fluents_inc_dec.add(effect.fluent)
            else:
                raise NotImplementedError
        if simulated_effect is not None and effect.fluent in simulated_effect.fluents:
            raise UPConflictingEffectsException(
                f"The effect {effect} is in conflict with the simulated effects already in the action."
            )
        self._effects.setdefault(timing, []).append(effect)

    @property
    def simulated_effects(
        self,
    ) -> Dict["up.model.timing.Timing", "up.model.effect.SimulatedEffect"]:
        """Returns the action simulated effects."""
        return self._simulated_effects

    def set_simulated_effect(
        self,
        timing: "up.model.timing.Timing",
        simulated_effect: "up.model.effect.SimulatedEffect",
    ):
        """Sets the given simulated effect at the specified timing"""
        for f in simulated_effect.fluents:
            if f in self._fluents_assigned.get(
                timing, set()
            ) or f in self._fluents_inc_dec.get(timing, set()):
                raise UPConflictingEffectsException(
                    f"The simulated effect {simulated_effect} is in conflict with the effects already in the action."
                )
        self._simulated_effects[timing] = simulated_effect



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
from typing import List, Union, Optional

from unified_planning.environment import get_env, Environment
from unified_planning.model.timing import Timing
from unified_planning.model.parameter import Parameter
from unified_planning.model.fnode import FNode
from unified_planning.model.types import Type
from unified_planning.model.expression import Expression
from unified_planning.model.operators import OperatorKind
from unified_planning.model.action import Action
from unified_planning.model.htn.task import Task, Subtask
from unified_planning.model.timing import Timepoint
from unified_planning.model.walkers import OperatorsExtractor


class TaskNetwork:
    """Represents an initial task network."""

    def __init__(self, _env: Environment = None):
        self._env = get_env(_env)
        self._variables: OrderedDict[str, Parameter] = OrderedDict()
        self._subtasks: List[Subtask] = []
        self._constraints: List[FNode] = []
        self._operators_extractor = OperatorsExtractor()  # maybe add to Environment?

    def __repr__(self):
        s = ["task network {\n"]
        if len(self._variables) > 0:
            s.append("  variables = [\n")
            for v in self.variables:
                s.append(f"    {v}\n")
            s.append("  ]\n")
        s.append("  subtasks = [\n")
        for t in self.subtasks:
            s.append(f"    {t}\n")
        s.append("  ]\n")
        if len(self._constraints) > 0:
            s.append("  constraints = [\n")
            for c in self.constraints:
                s.append(f"    {c}\n")
            s.append("  ]\n")
        s.append("}")
        return "".join(s)

    def __eq__(self, oth):
        if not isinstance(oth, TaskNetwork):
            return False
        return (
            set(self.variables) == set(oth.variables)
            and set(self.subtasks) == set(oth.subtasks)
            and set(self.constraints) == set(oth.constraints)
        )

    def __hash__(self):
        return (
            sum(map(hash, self.variables))
            + sum(map(hash, self.subtasks))
            + sum(map(hash, self.constraints))
        )

    def clone(self):
        new = TaskNetwork(self._env)
        new._variables: OrderedDict[str, Parameter] = self._variables.copy()
        new._subtasks: List[Subtask] = self._subtasks[:]
        new._constraints: List[FNode] = self._constraints[:]
        return new

    @property
    def variables(self) -> List[Parameter]:
        return list(self._variables.values())

    def add_variable(self, name: str, typename: Type) -> Parameter:
        assert (
            name not in self._variables
        ), f"A variable with name {name} already exists."
        param = Parameter(name, typename, self._env)
        self._variables[name] = param
        return param

    @property
    def subtasks(self) -> List["Subtask"]:
        """Returns the list of the subtasks."""
        return self._subtasks

    def add_subtask(
        self,
        task: Union[Subtask, Action, Task],
        *args: Expression,
        ident: Optional[str] = None,
    ) -> Subtask:
        if isinstance(task, Subtask):
            assert len(args) == 0 and ident is None
            subtask = task
        else:
            subtask = Subtask(task, *args, ident=ident)
        assert all([subtask.identifier != prev.identifier for prev in self.subtasks])
        self._subtasks.append(subtask)
        return subtask

    @property
    def constraints(self) -> List[FNode]:
        return self._constraints

    def add_constraint(self, constraint: Expression):
        (constraint,) = self._env.expression_manager.auto_promote(constraint)
        assert isinstance(constraint, FNode)
        assert self._env.type_checker.get_type(constraint).is_bool_type()
        assert OperatorKind.FLUENT_EXP not in self._operators_extractor.get(
            constraint
        ), f"The expression is not static (references a fluent): {constraint}"
        self._constraints.append(constraint)

    def set_strictly_before(
        self,
        lhs: Union[Subtask, Timepoint, Timing],
        rhs: Union[Subtask, Timepoint, Timing],
    ):
        if isinstance(lhs, Subtask):
            lhs = lhs.end
        if isinstance(lhs, Timepoint):
            lhs = Timing(timepoint=lhs, delay=0)
        assert isinstance(lhs, Timing)
        if isinstance(rhs, Subtask):
            rhs = rhs.start
        if isinstance(rhs, Timepoint):
            rhs = Timing(timepoint=rhs, delay=0)
        assert isinstance(rhs, Timing)
        self.add_constraint(self._env.expression_manager.LT(lhs, rhs))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.environment import get_env
from unified_planning.model.walkers import Dnf, Nnf, Substituter


class TestDnf(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.sub = Substituter(get_env())
        self.simp = get_env().simplifier

    def _subs_simp(self, exp, subs):
        ne = self.sub.substitute(exp, subs)
        return self.simp.simplify(ne)

    def test_nnf_dnf_1(self):
        n = Nnf(get_env())
        dnf = Dnf(get_env())

        a = FluentExp(Fluent("a"))
        b = FluentExp(Fluent("b"))
        c = FluentExp(Fluent("c"))
        # !(a => (b && c))
        e1 = Not(Implies(a, And(b, c)))
        nnf1 = n.get_nnf_expression(e1)
        self.assertIn("(a and ((not b) or (not c)))", str(nnf1))
        dnf1 = dnf.get_dnf_expression(e1)
        self.assertIn("((a and (not b)) or (a and (not c)))", str(dnf1))

    def test_dnf_2(self):
        dnf = Dnf(get_env())

        a = FluentExp(Fluent("a"))
        b = FluentExp(Fluent("b"))
        c = FluentExp(Fluent("c"))
        d = FluentExp(Fluent("d"))
        # a && (!b || (!c && d))
        e2 = And(a, Or(Not(b), And(Not(c), d)))
        dnf2 = dnf.get_dnf_expression(e2)
        self.assertIn("((a and (not b)) or (a and (not c) and d))", str(dnf2))

    def test_nnf_dnf_3(self):
        n = Nnf(get_env())
        dnf = Dnf(get_env())

        a = FluentExp(Fluent("a"))
        b = FluentExp(Fluent("b"))
        c = FluentExp(Fluent("c"))
        # (a => b) Iff (a => c)
        e3 = Iff(Implies(a, b), Implies(a, c))
        nnf3 = n.get_nnf_expression(e3)
        dnf3 = dnf.get_dnf_expression(e3)
        self.assertIn(
            "(((not a) or b) and ((not a) or c)) or ((a and (not b)) and (a and (not c)))",
            str(nnf3),
        )
        self.assertIn(
            "(not a) or ((not a) and c) or (b and (not a)) or (b and c) or (a and (not b) and (not c)",
            str(dnf3),
        )

    def test_nnf_dnf_4(self):
        n = Nnf(get_env())
        dnf = Dnf(get_env())

        a = FluentExp(Fluent("a"))
        b = FluentExp(Fluent("b"))
        # (a && ( a => b)) Iff (b || ( ((a => b ) && (b => a)) Iff ( a Iff b)))
        e4 = Iff(
            And(a, Implies(a, b)),
            Or(b, Iff(And(Implies(a, b), Implies(b, a)), Iff(a, b))),
        )
        nnf4 = n.get_nnf_expression(e4)
        dnf4 = dnf.get_dnf_expression(e4)
        subs = {a: True, b: True}
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(nnf4, subs))
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(dnf4, subs))
        subs = {a: True, b: False}
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(nnf4, subs))
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(dnf4, subs))
        subs = {a: False, b: False}
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(nnf4, subs))
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(dnf4, subs))
        subs = {a: False, b: True}
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(nnf4, subs))
        self.assertEqual(self._subs_simp(e4, subs), self._subs_simp(dnf4, subs))

    def test_nnf_dnf_5(self):
        n = Nnf(get_env())
        dnf = Dnf(get_env())

        a = FluentExp(Fluent("a"))
        b = FluentExp(Fluent("b"))
        c = FluentExp(Fluent("c"))
        d = FluentExp(Fluent("d"))
        # ((a && (c => a)) => d) Iff (( b => d) => c)
        e5 = Iff(Implies(And(a, Implies(c, a)), d), Implies(Implies(b, d), c))
        nnf5 = n.get_nnf_expression(e5)
        dnf5 = dnf.get_dnf_expression(e5)
        subs = {a: False, b: False, c: False, d: False}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: False, c: False, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: False, c: True, d: False}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: False, c: True, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: True, c: False, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: True, c: False, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: True, c: True, d: False}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: False, b: True, c: True, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: False, c: False, d: False}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: False, c: False, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: False, c: True, d: False}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: False, c: True, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: True, c: False, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: True, c: False, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: True, c: True, d: False}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))
        subs = {a: True, b: True, c: True, d: True}
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(nnf5, subs))
        self.assertEqual(self._subs_simp(e5, subs), self._subs_simp(dnf5, subs))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from collections import namedtuple

Example = namedtuple("Example", ["problem", "plan"])


def get_example_problems():
    problems = {}

    # basic
    x = Fluent("x")
    a = InstantaneousAction("a")
    a.add_precondition(Not(x))
    a.add_effect(x, True)
    problem = Problem("basic")
    problem.add_fluent(x)
    problem.add_action(a)
    problem.set_initial_value(x, False)
    problem.add_goal(x)
    plan = up.plans.SequentialPlan([up.plans.ActionInstance(a)])
    basic = Example(problem=problem, plan=plan)
    problems["basic"] = basic

    # basic conditional
    x = Fluent("x")
    y = Fluent("y")
    a_x = InstantaneousAction("a_x")
    a_y = InstantaneousAction("a_y")
    a_x.add_precondition(Not(x))
    a_x.add_effect(x, True, y)
    a_y.add_precondition(Not(y))
    a_y.add_effect(y, True)
    problem = Problem("basic_conditional")
    problem.add_fluent(x)
    problem.add_fluent(y)
    problem.add_action(a_x)
    problem.add_action(a_y)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y, False)
    problem.add_goal(x)
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(a_y), up.plans.ActionInstance(a_x)]
    )
    basic_conditional = Example(problem=problem, plan=plan)
    problems["basic_conditional"] = basic_conditional

    # basic oversubscription
    x = Fluent("x")
    a = InstantaneousAction("a")
    a.add_precondition(Not(x))
    a.add_effect(x, True)
    problem = Problem("basic_oversubscription")
    problem.add_fluent(x)
    problem.add_action(a)
    problem.set_initial_value(x, False)
    qm = up.model.metrics.Oversubscription({FluentExp(x): 10})
    problem.add_quality_metric(qm)
    plan = up.plans.SequentialPlan([up.plans.ActionInstance(a)])
    basic_oversubscription = Example(problem=problem, plan=plan)
    problems["basic_oversubscription"] = basic_oversubscription

    # complex conditional
    a = Fluent("a")
    b = Fluent("b")
    c = Fluent("c")
    d = Fluent("d")
    k = Fluent("k")
    x = Fluent("x")
    y = Fluent("y")
    z = Fluent("z")
    a_act = InstantaneousAction("act")
    a_0_act = InstantaneousAction("act_0")
    a_1_act = InstantaneousAction("act_1")
    a_2_act = InstantaneousAction("act_2")
    a_act.add_precondition(Not(a))
    a_act.add_effect(a, TRUE())
    a_act.add_effect(k, TRUE(), b)
    a_act.add_effect(x, TRUE(), Not(c))
    a_act.add_effect(y, FALSE(), d)
    a_0_act.add_precondition(Not(a))
    a_0_act.add_precondition(d)
    a_0_act.add_effect(b, TRUE())
    a_1_act.add_precondition(Not(a))
    a_1_act.add_precondition(d)
    a_1_act.add_precondition(b)
    a_1_act.add_effect(c, FALSE(), c)
    a_1_act.add_effect(c, TRUE(), Not(c))
    a_2_act.add_effect(a, FALSE())
    a_2_act.add_effect(d, TRUE())
    a_2_act.add_effect(z, FALSE(), z)
    a_2_act.add_effect(z, TRUE(), Not(z))
    problem = Problem("complex_conditional")
    problem.add_fluent(a)
    problem.add_fluent(b)
    problem.add_fluent(c)
    problem.add_fluent(d)
    problem.add_fluent(k)
    problem.add_fluent(x)
    problem.add_fluent(y)
    problem.add_fluent(z)
    problem.add_action(a_act)
    problem.add_action(a_0_act)
    problem.add_action(a_1_act)
    problem.add_action(a_2_act)
    problem.set_initial_value(a, True)
    problem.set_initial_value(b, False)
    problem.set_initial_value(c, True)
    problem.set_initial_value(d, False)
    problem.set_initial_value(k, False)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y, True)
    problem.set_initial_value(z, False)
    problem.add_goal(a)
    problem.add_goal(b)
    problem.add_goal(Not(c))
    problem.add_goal(d)
    problem.add_goal(k)
    problem.add_goal(x)
    problem.add_goal(Not(y))
    problem.add_goal(z)
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(a_2_act),
            up.plans.ActionInstance(a_0_act),
            up.plans.ActionInstance(a_1_act),
            up.plans.ActionInstance(a_act),
        ]
    )
    complex_conditional = Example(problem=problem, plan=plan)
    problems["complex_conditional"] = complex_conditional

    # basic without negative preconditions
    x = Fluent("x")
    y = Fluent("y")
    a = InstantaneousAction("a")
    a.add_precondition(y)
    a.add_effect(x, True)
    problem = Problem("basic_without_negative_preconditions")
    problem.add_fluent(x)
    problem.add_fluent(y)
    problem.add_action(a)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y, True)
    problem.add_goal(x)
    plan = up.plans.SequentialPlan([up.plans.ActionInstance(a)])
    basic_without_negative_preconditions = Example(problem=problem, plan=plan)
    problems[
        "basic_without_negative_preconditions"
    ] = basic_without_negative_preconditions

    # basic nested conjunctions
    problem = Problem("basic_nested_conjunctions")
    x = problem.add_fluent("x")
    y = problem.add_fluent("y")
    z = problem.add_fluent("z")
    j = problem.add_fluent("j")
    k = problem.add_fluent("k")

    a = InstantaneousAction("a")
    a.add_precondition(And(y, And(z, j, k)))
    a.add_effect(x, True)
    problem.add_action(a)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y, True)
    problem.set_initial_value(z, True)
    problem.set_initial_value(j, True)
    problem.set_initial_value(k, True)
    problem.add_goal(And(x, And(y, z, And(j, k))))
    plan = up.plans.SequentialPlan([up.plans.ActionInstance(a)])
    basic_nested_conjunctions = Example(problem=problem, plan=plan)
    problems["basic_nested_conjunctions"] = basic_nested_conjunctions

    # basic exists
    sem = UserType("Semaphore")
    x = Fluent("x")
    y = Fluent("y", BoolType(), semaphore=sem)
    o1 = Object("o1", sem)
    o2 = Object("o2", sem)
    s_var = Variable("s", sem)
    a = InstantaneousAction("a")
    a.add_precondition(Exists(FluentExp(y, [s_var]), s_var))
    a.add_effect(x, True)
    problem = Problem("basic_exists")
    problem.add_fluent(x)
    problem.add_fluent(y)
    problem.add_object(o1)
    problem.add_object(o2)
    problem.add_action(a)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y(o1), True)
    problem.set_initial_value(y(o2), False)
    problem.add_goal(x)
    plan = up.plans.SequentialPlan([up.plans.ActionInstance(a)])
    basic_exists = Example(problem=problem, plan=plan)
    problems["basic_exists"] = basic_exists

    # basic forall
    sem = UserType("Semaphore")
    x = Fluent("x")
    y = Fluent("y", BoolType(), semaphore=sem)
    s_var = Variable("s", sem)
    a = InstantaneousAction("a")
    a.add_precondition(Forall(Not(y(s_var)), s_var))
    a.add_effect(x, True)
    problem = Problem("basic_forall")
    problem.add_fluent(x)
    problem.add_fluent(y)
    o1 = problem.add_object("o1", sem)
    o2 = problem.add_object("o2", sem)
    problem.add_action(a)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y(o1), False)
    problem.set_initial_value(y(o2), False)
    problem.add_goal(x)
    plan = up.plans.SequentialPlan([up.plans.ActionInstance(a)])
    basic_forall = Example(problem=problem, plan=plan)
    problems["basic_forall"] = basic_forall

    # temporal conditional
    Obj = UserType("Obj")
    is_same_obj = Fluent("is_same_obj", BoolType(), object_1=Obj, object_2=Obj)
    is_ok = Fluent("is_ok", BoolType(), test=Obj)
    is_ok_giver = Fluent("is_ok_giver", BoolType(), test=Obj)
    ok_given = Fluent("ok_given")
    set_giver = DurativeAction("set_giver", y=Obj)
    y = set_giver.parameter("y")
    set_giver.set_fixed_duration(2)
    set_giver.add_condition(StartTiming(), Not(is_ok_giver(y)))
    set_giver.add_effect(StartTiming(), is_ok_giver(y), True)
    set_giver.add_effect(EndTiming(), is_ok_giver(y), False)
    take_ok = DurativeAction("take_ok", x=Obj, y=Obj)
    x = take_ok.parameter("x")
    y = take_ok.parameter("y")
    take_ok.set_fixed_duration(3)
    take_ok.add_condition(StartTiming(), Not(is_ok(x)))
    take_ok.add_condition(StartTiming(), Not(is_ok_giver(y)))
    take_ok.add_condition(StartTiming(), Not(is_same_obj(x, y)))
    take_ok.add_effect(EndTiming(), is_ok(x), True, is_ok_giver(y))
    take_ok.add_effect(EndTiming(), ok_given, True)
    o1 = Object("o1", Obj)
    o2 = Object("o2", Obj)
    problem = Problem("temporal_conditional")
    problem.add_fluent(is_same_obj, default_initial_value=False)
    problem.add_fluent(is_ok, default_initial_value=False)
    problem.add_fluent(is_ok_giver, default_initial_value=False)
    problem.add_fluent(ok_given, default_initial_value=False)
    problem.add_action(set_giver)
    problem.add_action(take_ok)
    problem.add_object(o1)
    problem.add_object(o2)
    problem.add_goal(is_ok(o1))
    problem.set_initial_value(is_same_obj(o1, o1), True)
    problem.set_initial_value(is_same_obj(o2, o2), True)
    plan = up.plans.TimeTriggeredPlan(
        [
            (
                Fraction(0, 1),
                up.plans.ActionInstance(take_ok, (ObjectExp(o1), ObjectExp(o2))),
                Fraction(3, 1),
            ),
            (
                Fraction(1, 1),
                up.plans.ActionInstance(set_giver, (ObjectExp(o2),)),
                Fraction(2, 1),
            ),
        ]
    )
    temporal_conditional = Example(problem=problem, plan=plan)
    problems["temporal_conditional"] = temporal_conditional

    # basic with actions cost
    x = Fluent("x")
    y = Fluent("y")
    a = InstantaneousAction("a")
    a.add_precondition(Not(x))
    a.add_effect(x, True)
    b = InstantaneousAction("b")
    b.add_precondition(Not(y))
    b.add_effect(y, True)
    c = InstantaneousAction("c")
    c.add_precondition(y)
    c.add_effect(x, True)
    problem = Problem("basic_with_costs")
    problem.add_fluent(x)
    problem.add_fluent(y)
    problem.add_action(a)
    problem.add_action(b)
    problem.add_action(c)
    problem.set_initial_value(x, False)
    problem.set_initial_value(y, False)
    problem.add_goal(x)
    problem.add_quality_metric(
        up.model.metrics.MinimizeActionCosts({a: Int(10), b: Int(1), c: Int(1)})
    )
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(b), up.plans.ActionInstance(c)]
    )
    basic_with_costs = Example(problem=problem, plan=plan)
    problems["basic_with_costs"] = basic_with_costs

    # counter
    counter_1 = Fluent("counter_1", IntType(0, 10))
    counter_2 = Fluent("counter_2", IntType(0, 10))
    fake_counter = Fluent("fake_counter", IntType(0, 10))
    increase = InstantaneousAction("increase")
    increase.add_increase_effect(counter_1, 1)
    increase.add_effect(counter_2, Plus(counter_2, 1))
    increase.add_effect(fake_counter, Div(Times(fake_counter, 2), 2))
    problem = Problem("counter")
    problem.add_fluent(counter_1)
    problem.add_fluent(counter_2)
    problem.add_fluent(fake_counter)
    problem.add_action(increase)
    problem.set_initial_value(counter_1, 0)
    problem.set_initial_value(counter_2, 0)
    problem.set_initial_value(fake_counter, 1)
    problem.add_goal(Iff(LT(fake_counter, counter_1), LT(counter_2, 3)))
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(increase), up.plans.ActionInstance(increase)]
    )
    counter = Example(problem=problem, plan=plan)
    problems["counter"] = counter

    # counter to 50
    counter = Fluent("counter", IntType(0, 100))
    increase = InstantaneousAction("increase")
    increase.add_increase_effect(counter, 1)
    problem = Problem("counter_to_50")
    problem.add_fluent(counter)
    problem.add_action(increase)
    problem.set_initial_value(counter, 0)
    problem.add_goal(Equals(counter, 50))
    # Make a plan of 50 action instances of "increase"
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(increase) for _ in range(50)]
    )
    counter_to_50 = Example(problem=problem, plan=plan)
    problems["counter_to_50"] = counter_to_50

    # basic with object constant
    Location = UserType("Location")
    is_at = Fluent("is_at", BoolType(), loc=Location)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(is_at(l_from))
    move.add_precondition(Not(is_at(l_to)))
    move.add_effect(is_at(l_from), False)
    move.add_effect(is_at(l_to), True)
    move_to_l1 = InstantaneousAction("move_to_l1", l_from=Location)
    l_from = move_to_l1.parameter("l_from")
    move_to_l1.add_precondition(is_at(l_from))
    move_to_l1.add_precondition(Not(is_at(l1)))
    move_to_l1.add_effect(is_at(l_from), False)
    move_to_l1.add_effect(is_at(l1), True)
    problem = Problem("basic_with_object_constant")
    problem.add_fluent(is_at)
    problem.add_objects([l1, l2])
    problem.add_action(move)
    problem.add_action(move_to_l1)
    problem.set_initial_value(is_at(l1), True)
    problem.set_initial_value(is_at(l2), False)
    problem.add_goal(is_at(l2))
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))]
    )
    basic_with_object_constant = Example(problem=problem, plan=plan)
    problems["basic_with_object_constant"] = basic_with_object_constant

    return problems



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, Iterator, List, Optional, Set, Tuple, Union, cast
import unified_planning as up
from unified_planning.engines.compilers import Grounder
from unified_planning.engines.engine import Engine
from unified_planning.engines.mixins.simulator import Event, SimulatorMixin
from unified_planning.exceptions import UPUsageError, UPConflictingEffectsException
from unified_planning.plans import ActionInstance
from unified_planning.model.walkers import StateEvaluator


class InstantaneousEvent(Event):
    """Implements the Event class for an Instantaneous Action."""

    def __init__(
        self,
        conditions: List["up.model.FNode"],
        effects: List["up.model.Effect"],
        simulated_effect: Optional["up.model.SimulatedEffect"] = None,
    ):
        self._conditions = conditions
        self._effects = effects
        self._simulated_effect = simulated_effect

    @property
    def conditions(self) -> List["up.model.FNode"]:
        return self._conditions

    @property
    def effects(self) -> List["up.model.Effect"]:
        return self._effects

    @property
    def simulated_effect(self) -> Optional["up.model.SimulatedEffect"]:
        return self._simulated_effect


class SequentialSimulator(Engine, SimulatorMixin):
    """
    Sequential SimulatorMixin implementation.
    """

    def __init__(self, problem: "up.model.Problem"):
        SimulatorMixin.__init__(self, problem)
        pk = problem.kind
        assert Grounder.supports(pk)
        assert isinstance(self._problem, up.model.Problem)
        grounder = Grounder()
        self._grounding_result = grounder.compile(
            self._problem, up.engines.CompilationKind.GROUNDING
        )

        grounded_problem: "up.model.Problem" = cast(
            up.model.Problem, self._grounding_result.problem
        )
        lift_map = self._grounding_result.map_back_action_instance
        assert lift_map is not None
        self._events: Dict[
            Tuple["up.model.Action", Tuple["up.model.FNode", ...]], List[Event]
        ] = {}
        for grounded_action in grounded_problem.actions:
            if isinstance(grounded_action, up.model.InstantaneousAction):
                lifted_ai = lift_map(ActionInstance(grounded_action))
                event_list = self._events.setdefault(
                    (lifted_ai.action, lifted_ai.actual_parameters), []
                )
                event_list.append(
                    InstantaneousEvent(
                        grounded_action.preconditions,
                        grounded_action.effects,
                        grounded_action.simulated_effect,
                    )
                )
            else:
                raise NotImplementedError
        self._se = StateEvaluator(self._problem)

    def _get_unsatisfied_conditions(
        self, event: "Event", state: "up.model.ROState", early_termination: bool = False
    ) -> List["up.model.FNode"]:
        """
        Returns the list of unsatisfied event conditions evaluated in the given state.
        If the flag "early_termination" is set, the method ends and returns at the first unsatisfied condition.
        :param state: The State in which the event conditions are evaluated.
        :param early_termination: Flag deciding if the method ends and returns at the first unsatisfied condition.
        :return: The list of all the event conditions that evaluated to False or the list containing the first condition evaluated to False if the flag "early_termination" is set.
        """
        # Evaluate every condition and if the condition is False or the condition is not simplified as a
        # boolean constant in the given state, return False. Return True otherwise
        unsatisfied_conditions = []
        for c in event.conditions:
            evaluated_cond = self._se.evaluate(c, state)
            if (
                not evaluated_cond.is_bool_constant()
                or not evaluated_cond.bool_constant_value()
            ):
                unsatisfied_conditions.append(c)
                if early_termination:
                    break
        return unsatisfied_conditions

    def _apply(
        self, event: "Event", state: "up.model.COWState"
    ) -> Optional["up.model.COWState"]:
        """
        Returns None if the event is not applicable in the given state, otherwise returns a new COWState,
        which is a copy of the given state but the applicable effects of the event are applied; therefore
        some fluent values are updated.
        :param state: the state where the event formulas are calculated.
        :param event: the event that has the information about the conditions to check and the effects to apply.
        :return: None if the event is not applicable in the given state, a new COWState with some updated values
         if the event is applicable.
        """
        if not self.is_applicable(event, state):
            return None
        else:
            return self.apply_unsafe(event, state)

    def _apply_unsafe(
        self, event: "Event", state: "up.model.COWState"
    ) -> "up.model.COWState":
        """
        Returns a new COWState, which is a copy of the given state but the applicable effects of the event are applied; therefore
        some fluent values are updated.
        IMPORTANT NOTE: Assumes that self.is_applicable(state, event) returns True
        :param state: the state where the event formulas are evaluated.
        :param event: the event that has the information about the effects to apply.
        :return: A new COWState with some updated values.
        """
        updated_values: Dict["up.model.FNode", "up.model.FNode"] = {}
        assigned_fluent: Set["up.model.FNode"] = set()
        em = self._problem.env.expression_manager
        for e in event.effects:
            evaluated_args = tuple(self._se.evaluate(a, state) for a in e.fluent.args)
            fluent = self._problem.env.expression_manager.FluentExp(
                e.fluent.fluent(), evaluated_args
            )
            if (not e.is_conditional()) or self._se.evaluate(
                e.condition, state
            ).is_true():
                if e.is_assignment():
                    if fluent in updated_values:
                        raise UPConflictingEffectsException(
                            f"The fluent {fluent} is modified by 2 assignments in the same event."
                        )
                    updated_values[fluent] = self._se.evaluate(e.value, state)
                    assigned_fluent.add(fluent)
                else:
                    if fluent in assigned_fluent:
                        raise UPConflictingEffectsException(
                            f"The fluent {fluent} is modified by an assignment and an increase/decrease in the same event."
                        )
                    # If the fluent is in updated_values, we take his modified value, (which was modified by another increase or deacrease)
                    # otherwisee we take it's evaluation in the state as it's value.
                    f_eval = updated_values.get(
                        fluent, self._se.evaluate(fluent, state)
                    )
                    v_eval = self._se.evaluate(e.value, state)
                    if e.is_increase():
                        updated_values[fluent] = em.auto_promote(
                            f_eval.constant_value() + v_eval.constant_value()
                        )[0]
                    elif e.is_decrease():
                        updated_values[fluent] = em.auto_promote(
                            f_eval.constant_value() - v_eval.constant_value()
                        )[0]
                    else:
                        raise NotImplementedError
        if event.simulated_effect is not None:
            for f, v in zip(
                event.simulated_effect.fluents,
                event.simulated_effect.function(self._problem, state, {}),
            ):
                if f in updated_values:
                    raise UPConflictingEffectsException(
                        f"The fluent {f} is modified twice in the same event."
                    )
                updated_values[f] = v
        return state.make_child(updated_values)

    def _get_applicable_events(self, state: "up.model.ROState") -> Iterator["Event"]:
        """
        Returns a view over all the events that are applicable in the given State;
        an Event is considered applicable in a given State, when all the Event condition
        simplify as True when evaluated in the State.
        :param state: the state where the formulas are evaluated.
        :return: an Iterator of applicable Events.
        """
        for events in self._events.values():
            for event in events:
                if self.is_applicable(event, state):
                    yield event

    def _get_events(
        self,
        action: "up.model.Action",
        parameters: Union[
            Tuple["up.model.Expression", ...], List["up.model.Expression"]
        ],
    ) -> List["Event"]:
        """
        Returns a list containing all the events derived from the given action, grounded with the given parameters.
        :param action: the action containing the information to create the event.
        :param parameters: the parameters needed to ground the action
        :return: the List of Events derived from this action with these parameters.
        """
        if action not in cast(up.model.Problem, self._problem).actions:
            raise UPUsageError(
                "The action given as parameter does not belong to the problem given to the SequentialSimulator."
            )
        params_exp = tuple(
            self._problem.env.expression_manager.auto_promote(parameters)
        )
        return self._events.get((action, tuple(params_exp)), [])

    def _get_unsatisfied_goals(
        self, state: "up.model.ROState", early_termination: bool = False
    ) -> List["up.model.FNode"]:
        """
        Returns the list of unsatisfied goals evaluated in the given state.
        If the flag "early_termination" is set, the method ends and returns at the first unsatisfied goal.
        :param state: The State in which the problem goals are evaluated.
        :param early_termination: Flag deciding if the method ends and returns at the first unsatisfied goal.
        :return: The list of all the goals that evaluated to False or the list containing the first goal evaluated to False if the flag "early_termination" is set.
        """
        unsatisfied_goals = []
        for g in cast(up.model.Problem, self._problem).goals:
            g_eval = self._se.evaluate(g, state).bool_constant_value()
            if not g_eval:
                unsatisfied_goals.append(g)
                if early_termination:
                    break
        return unsatisfied_goals

    @property
    def name(self) -> str:
        return "sequential_simulator"

    @staticmethod
    def supported_kind() -> "up.model.ProblemKind":
        supported_kind = up.model.ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_simulated_entities("SIMULATED_EFFECTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= SequentialSimulator.supported_kind()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from collections import namedtuple

Example = namedtuple("Example", ["problem", "plan"])


def get_example_problems():
    problems = {}

    # robot
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    battery_charge = Fluent("battery_charge", RealType(0, 100))
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(GE(battery_charge, 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_effect(battery_charge, Minus(battery_charge, 10))
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot")
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100)
    problem.add_goal(robot_at(l2))
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))]
    )
    robot = Example(problem=problem, plan=plan)
    problems["robot"] = robot

    # robot fluent of user_type
    Location = UserType("Location")
    Robot = UserType("Robot")
    is_at = Fluent("is_at", Location, robot=Robot)
    move = InstantaneousAction("move", robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter("robot")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(Equals(is_at(robot), l_from))
    move.add_precondition(Not(Equals(is_at(robot), l_to)))
    move.add_effect(is_at(robot), l_to)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    r1 = Object("r1", Robot)
    r2 = Object("r2", Robot)
    problem = Problem("robot_fluent_of_user_type")
    problem.add_fluent(is_at)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(r1)
    problem.add_object(r2)
    problem.set_initial_value(is_at(r1), l2)
    problem.set_initial_value(is_at(r2), l1)
    problem.add_goal(Equals(is_at(r1), l1))
    problem.add_goal(Equals(is_at(r2), l2))
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(
                move, (ObjectExp(r1), ObjectExp(l2), ObjectExp(l1))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(r2), ObjectExp(l1), ObjectExp(l2))
            ),
        ]
    )
    robot_fluent_of_user_type = Example(problem=problem, plan=plan)
    problems["robot_fluent_of_user_type"] = robot_fluent_of_user_type

    # robot no negative preconditions
    Location = UserType("location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(robot_at(l_from))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot")
    problem.add_fluent(robot_at)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.add_goal(robot_at(l2))
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))]
    )
    robot_no_negative_preconditions = Example(problem=problem, plan=plan)
    problems["robot_no_negative_preconditions"] = robot_no_negative_preconditions

    # robot decrease
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    battery_charge = Fluent("battery_charge", RealType(0, 100))
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(GE(battery_charge, 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_decrease_effect(battery_charge, 10)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot_decrease")
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100)
    problem.add_goal(robot_at(l2))
    plan = up.plans.SequentialPlan(
        [up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))]
    )
    robot_decrease = Example(problem=problem, plan=plan)
    problems["robot_decrease"] = robot_decrease

    # robot_loader
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    cargo_at = Fluent("cargo_at", BoolType(), position=Location)
    cargo_mounted = Fluent("cargo_mounted")
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    load = InstantaneousAction("load", loc=Location)
    loc = load.parameter("loc")
    load.add_precondition(cargo_at(loc))
    load.add_precondition(robot_at(loc))
    load.add_precondition(Not(cargo_mounted))
    load.add_effect(cargo_at(loc), False)
    load.add_effect(cargo_mounted, True)
    unload = InstantaneousAction("unload", loc=Location)
    loc = unload.parameter("loc")
    unload.add_precondition(Not(cargo_at(loc)))
    unload.add_precondition(robot_at(loc))
    unload.add_precondition(cargo_mounted)
    unload.add_effect(cargo_at(loc), True)
    unload.add_effect(cargo_mounted, False)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot_loader")
    problem.add_fluent(robot_at)
    problem.add_fluent(cargo_at)
    problem.add_fluent(cargo_mounted)
    problem.add_action(move)
    problem.add_action(load)
    problem.add_action(unload)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(cargo_at(l1), False)
    problem.set_initial_value(cargo_at(l2), True)
    problem.set_initial_value(cargo_mounted, False)
    problem.add_goal(cargo_at(l1))
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2))),
            up.plans.ActionInstance(load, (ObjectExp(l2),)),
            up.plans.ActionInstance(move, (ObjectExp(l2), ObjectExp(l1))),
            up.plans.ActionInstance(unload, (ObjectExp(l1),)),
        ]
    )
    robot_loader = Example(problem=problem, plan=plan)
    problems["robot_loader"] = robot_loader

    # robot_loader_mod
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    cargo_at = Fluent("cargo_at", BoolType(), position=Location)
    is_same_location = Fluent("is_same_location", BoolType(), p1=Location, p2=Location)
    cargo_mounted = Fluent("cargo_mounted")
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_precondition(Not(is_same_location(l_from, l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    load = InstantaneousAction("load", loc=Location)
    loc = load.parameter("loc")
    load.add_precondition(cargo_at(loc))
    load.add_precondition(robot_at(loc))
    load.add_precondition(Not(cargo_mounted))
    load.add_effect(cargo_at(loc), False)
    load.add_effect(cargo_mounted, True)
    unload = InstantaneousAction("unload", loc=Location)
    loc = unload.parameter("loc")
    unload.add_precondition(Not(cargo_at(loc)))
    unload.add_precondition(robot_at(loc))
    unload.add_precondition(cargo_mounted)
    unload.add_effect(cargo_at(loc), True)
    unload.add_effect(cargo_mounted, False)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot_loader_mod")
    problem.add_fluent(robot_at, default_initial_value=False)
    problem.add_fluent(cargo_at, default_initial_value=False)
    problem.add_fluent(cargo_mounted, default_initial_value=False)
    problem.add_fluent(is_same_location, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(load)
    problem.add_action(unload)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(cargo_at(l2), True)
    for o in problem.objects(Location):
        problem.set_initial_value(is_same_location(o, o), True)
    problem.add_goal(cargo_at(l1))
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2))),
            up.plans.ActionInstance(load, (ObjectExp(l2),)),
            up.plans.ActionInstance(move, (ObjectExp(l2), ObjectExp(l1))),
            up.plans.ActionInstance(unload, (ObjectExp(l1),)),
        ]
    )
    robot_loader_mod = Example(problem=problem, plan=plan)
    problems["robot_loader_mod"] = robot_loader_mod

    # robot_loader_adv
    Robot = UserType("Robot")
    Container = UserType("Container")
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), robot=Robot, position=Location)
    cargo_at = Fluent("cargo_at", BoolType(), cargo=Container, position=Location)
    cargo_mounted = Fluent("cargo_mounted", BoolType(), cargo=Container, robot=Robot)
    move = InstantaneousAction("move", l_from=Location, l_to=Location, r=Robot)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    r = move.parameter("r")
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(r, l_from))
    move.add_precondition(Not(robot_at(r, l_to)))
    move.add_effect(robot_at(r, l_from), False)
    move.add_effect(robot_at(r, l_to), True)
    load = InstantaneousAction("load", loc=Location, r=Robot, c=Container)
    loc = load.parameter("loc")
    r = load.parameter("r")
    c = load.parameter("c")
    load.add_precondition(cargo_at(c, loc))
    load.add_precondition(robot_at(r, loc))
    load.add_precondition(Not(cargo_mounted(c, r)))
    load.add_effect(cargo_at(c, loc), False)
    load.add_effect(cargo_mounted(c, r), True)
    unload = InstantaneousAction("unload", loc=Location, r=Robot, c=Container)
    loc = unload.parameter("loc")
    r = unload.parameter("r")
    c = unload.parameter("c")
    unload.add_precondition(Not(cargo_at(c, loc)))
    unload.add_precondition(robot_at(r, loc))
    unload.add_precondition(cargo_mounted(c, r))
    unload.add_effect(cargo_at(c, loc), True)
    unload.add_effect(cargo_mounted(c, r), False)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    l3 = Object("l3", Location)
    r1 = Object("r1", Robot)
    c1 = Object("c1", Container)
    problem = Problem("robot_loader_adv")
    problem.add_fluent(robot_at)
    problem.add_fluent(cargo_at)
    problem.add_fluent(cargo_mounted)
    problem.add_action(move)
    problem.add_action(load)
    problem.add_action(unload)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(r1)
    problem.add_object(c1)
    problem.set_initial_value(robot_at(r1, l1), True)
    problem.set_initial_value(robot_at(r1, l2), False)
    problem.set_initial_value(robot_at(r1, l3), False)
    problem.set_initial_value(cargo_at(c1, l1), False)
    problem.set_initial_value(cargo_at(c1, l2), True)
    problem.set_initial_value(cargo_at(c1, l3), False)
    problem.set_initial_value(cargo_mounted(c1, r1), False)
    problem.add_goal(cargo_at(c1, l3))
    problem.add_goal(robot_at(r1, l1))
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(
                move, (ObjectExp(l1), ObjectExp(l2), ObjectExp(r1))
            ),
            up.plans.ActionInstance(
                load, (ObjectExp(l2), ObjectExp(r1), ObjectExp(c1))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(l2), ObjectExp(l3), ObjectExp(r1))
            ),
            up.plans.ActionInstance(
                unload, (ObjectExp(l3), ObjectExp(r1), ObjectExp(c1))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(l3), ObjectExp(l1), ObjectExp(r1))
            ),
        ]
    )
    robot_loader_adv = Example(problem=problem, plan=plan)
    problems["robot_loader_adv"] = robot_loader_adv

    # robot locations connected
    Location = UserType("Location")
    Robot = UserType("Robot")
    is_at = Fluent("is_at", BoolType(), position=Location, robot=Robot)
    battery_charge = Fluent("battery_charge", RealType(0, 100), robot=Robot)
    is_connected = Fluent(
        "is_connected", BoolType(), location_1=Location, location_2=Location
    )
    move = InstantaneousAction("move", robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter("robot")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(GE(battery_charge(robot), 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(is_at(l_from, robot))
    move.add_precondition(Not(is_at(l_to, robot)))
    move.add_precondition(Or(is_connected(l_from, l_to), is_connected(l_to, l_from)))
    move.add_effect(is_at(l_from, robot), False)
    move.add_effect(is_at(l_to, robot), True)
    move.add_decrease_effect(battery_charge(robot), 10)
    move_2 = InstantaneousAction("move_2", robot=Robot, l_from=Location, l_to=Location)
    robot = move_2.parameter("robot")
    l_from = move_2.parameter("l_from")
    l_to = move_2.parameter("l_to")
    move_2.add_precondition(GE(battery_charge(robot), 15))
    move_2.add_precondition(Not(Equals(l_from, l_to)))
    move_2.add_precondition(is_at(l_from, robot))
    move_2.add_precondition(Not(is_at(l_to, robot)))
    mid_location = Variable("mid_loc", Location)
    # (E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move_2.add_precondition(
        Exists(
            And(
                Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                Or(
                    is_connected(l_from, mid_location),
                    is_connected(mid_location, l_from),
                ),
                Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to)),
            ),
            mid_location,
        )
    )
    move_2.add_effect(is_at(l_from, robot), False)
    move_2.add_effect(is_at(l_to, robot), True)
    move_2.add_decrease_effect(battery_charge(robot), 15)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    l3 = Object("l3", Location)
    l4 = Object("l4", Location)
    l5 = Object("l5", Location)
    r1 = Object("r1", Robot)
    problem = Problem("robot_locations_connected")
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(battery_charge)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(move_2)
    problem.add_object(r1)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.set_initial_value(battery_charge(r1), 100)
    problem.add_goal(is_at(l5, r1))
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(
                move_2, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l3))
            ),
            up.plans.ActionInstance(
                move_2, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l5))
            ),
        ]
    )
    robot_locations_connected = Example(problem=problem, plan=plan)
    problems["robot_locations_connected"] = robot_locations_connected

    # robot locations visited
    Location = UserType("Location")
    Robot = UserType("Robot")
    is_at = Fluent("is_at", BoolType(), position=Location, robot=Robot)
    battery_charge = Fluent("battery_charge", RealType(0, 100), robot=Robot)
    is_connected = Fluent(
        "is_connected", BoolType(), location_1=Location, location_2=Location
    )
    visited = Fluent("visited", BoolType(), target=Location)
    move = InstantaneousAction("move", robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter("robot")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(GE(battery_charge(robot), 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(is_at(l_from, robot))
    move.add_precondition(Not(is_at(l_to, robot)))
    move.add_precondition(Or(is_connected(l_from, l_to), is_connected(l_to, l_from)))
    move.add_effect(is_at(l_from, robot), False)
    move.add_effect(is_at(l_to, robot), True)
    move.add_effect(visited(l_to), True)
    move.add_decrease_effect(battery_charge(robot), 10)
    move_2 = InstantaneousAction("move_2", robot=Robot, l_from=Location, l_to=Location)
    robot = move_2.parameter("robot")
    l_from = move_2.parameter("l_from")
    l_to = move_2.parameter("l_to")
    move_2.add_precondition(GE(battery_charge(robot), 15))
    move_2.add_precondition(Not(Equals(l_from, l_to)))
    move_2.add_precondition(is_at(l_from, robot))
    move_2.add_precondition(Not(is_at(l_to, robot)))
    mid_location = Variable("mid_loc", Location)
    # (E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move_2.add_precondition(
        Exists(
            And(
                Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                Or(
                    is_connected(l_from, mid_location),
                    is_connected(mid_location, l_from),
                ),
                Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to)),
            ),
            mid_location,
        )
    )
    move_2.add_effect(is_at(l_from, robot), False)
    move_2.add_effect(is_at(l_to, robot), True)
    move_2.add_effect(visited(l_to), True)
    move_2.add_decrease_effect(battery_charge(robot), 15)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    l3 = Object("l3", Location)
    l4 = Object("l4", Location)
    l5 = Object("l5", Location)
    r1 = Object("r1", Robot)
    problem = Problem("robot_locations_visited")
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(battery_charge)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_fluent(visited, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(move_2)
    problem.add_object(r1)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(visited(l1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.set_initial_value(battery_charge(r1), 50)
    problem.add_goal(is_at(l5, r1))
    visited_location = Variable("visited_loc", Location)
    problem.add_goal(Forall(visited(visited_location), visited_location))
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(
                move, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l2))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(r1), ObjectExp(l2), ObjectExp(l3))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l4))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(r1), ObjectExp(l4), ObjectExp(l5))
            ),
        ]
    )
    robot_locations_visited = Example(problem=problem, plan=plan)
    problems["robot_locations_visited"] = robot_locations_visited

    # charger_discharger
    charger = Fluent("charger")
    b_1 = Fluent("b_1")
    b_2 = Fluent("b_2")
    b_3 = Fluent("b_3")
    charge = InstantaneousAction("charge")
    discharge = InstantaneousAction("discharge")
    charge.add_precondition(Not(charger))
    charge.add_effect(charger, True)
    # !(charger => (b_1 && b_2 && b_3)) in dnf:
    # (charger and !b_1 ) or (charger and !b_2) or (charger and !b_3)
    # which represents the charger is full and at least one battery is not
    discharge.add_precondition(Not(Implies(charger, And(b_1, b_2, b_3))))
    discharge.add_effect(charger, False)
    discharge.add_effect(b_1, True, Not(b_1))
    discharge.add_effect(b_2, True, And(b_1, Not(b_2)))
    discharge.add_effect(b_3, True, And(b_1, b_2, Not(b_3)))
    problem = Problem("charger_discharger")
    problem.add_fluent(charger)
    problem.add_fluent(b_1)
    problem.add_fluent(b_2)
    problem.add_fluent(b_3)
    problem.add_action(charge)
    problem.add_action(discharge)
    problem.set_initial_value(charger, False)
    problem.set_initial_value(b_1, False)
    problem.set_initial_value(b_2, False)
    problem.set_initial_value(b_3, False)
    problem.add_goal(b_1)
    problem.add_goal(b_2)
    problem.add_goal(b_3)
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(charge),
            up.plans.ActionInstance(discharge),
            up.plans.ActionInstance(charge),
            up.plans.ActionInstance(discharge),
            up.plans.ActionInstance(charge),
            up.plans.ActionInstance(discharge),
        ]
    )
    charge_discharge = Example(problem=problem, plan=plan)
    problems["charge_discharge"] = charge_discharge

    # matchcellar
    Match = UserType("Match")
    Fuse = UserType("Fuse")
    handfree = Fluent("handfree")
    light = Fluent("light")
    match_used = Fluent("match_used", BoolType(), match=Match)
    fuse_mended = Fluent("fuse_mended", BoolType(), fuse=Fuse)
    light_match = DurativeAction("light_match", m=Match)
    m = light_match.parameter("m")
    light_match.set_fixed_duration(6)
    light_match.add_condition(StartTiming(), Not(match_used(m)))
    light_match.add_effect(StartTiming(), match_used(m), True)
    light_match.add_effect(StartTiming(), light, True)
    light_match.add_effect(EndTiming(), light, False)
    mend_fuse = DurativeAction("mend_fuse", f=Fuse)
    f = mend_fuse.parameter("f")
    mend_fuse.set_fixed_duration(5)
    mend_fuse.add_condition(StartTiming(), handfree)
    mend_fuse.add_condition(ClosedTimeInterval(StartTiming(), EndTiming()), light)
    mend_fuse.add_effect(StartTiming(), handfree, False)
    mend_fuse.add_effect(EndTiming(), fuse_mended(f), True)
    mend_fuse.add_effect(EndTiming(), handfree, True)
    f1 = Object("f1", Fuse)
    f2 = Object("f2", Fuse)
    f3 = Object("f3", Fuse)
    m1 = Object("m1", Match)
    m2 = Object("m2", Match)
    m3 = Object("m3", Match)
    problem = Problem("MatchCellar")
    problem.add_fluent(handfree)
    problem.add_fluent(light)
    problem.add_fluent(match_used, default_initial_value=False)
    problem.add_fluent(fuse_mended, default_initial_value=False)
    problem.add_action(light_match)
    problem.add_action(mend_fuse)
    problem.add_object(f1)
    problem.add_object(f2)
    problem.add_object(f3)
    problem.add_object(m1)
    problem.add_object(m2)
    problem.add_object(m3)
    problem.set_initial_value(light, False)
    problem.set_initial_value(handfree, True)
    problem.add_goal(fuse_mended(f1))
    problem.add_goal(fuse_mended(f2))
    problem.add_goal(fuse_mended(f3))
    plan = up.plans.TimeTriggeredPlan(
        [
            (
                Fraction(0, 1),
                up.plans.ActionInstance(light_match, (ObjectExp(m1),)),
                Fraction(6, 1),
            ),
            (
                Fraction(1, 100),
                up.plans.ActionInstance(mend_fuse, (ObjectExp(f1),)),
                Fraction(5, 1),
            ),
            (
                Fraction(601, 100),
                up.plans.ActionInstance(light_match, (ObjectExp(m2),)),
                Fraction(6, 1),
            ),
            (
                Fraction(602, 100),
                up.plans.ActionInstance(mend_fuse, (ObjectExp(f2),)),
                Fraction(5, 1),
            ),
            (
                Fraction(1202, 100),
                up.plans.ActionInstance(light_match, (ObjectExp(m3),)),
                Fraction(6, 1),
            ),
            (
                Fraction(1203, 100),
                up.plans.ActionInstance(mend_fuse, (ObjectExp(f3),)),
                Fraction(5, 1),
            ),
        ]
    )
    matchcellar = Example(problem=problem, plan=plan)
    problems["matchcellar"] = matchcellar

    # timed connected locations
    Location = UserType("Location")
    is_connected = Fluent(
        "is_connected", BoolType(), location_1=Location, location_2=Location
    )
    is_at = Fluent("is_at", BoolType(), position=Location)
    move = DurativeAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.set_fixed_duration(6)
    move.add_condition(StartTiming(), is_at(l_from))
    move.add_condition(StartTiming(), Not(is_at(l_to)))
    mid_location = Variable("mid_loc", Location)
    # (E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move.add_condition(
        ClosedTimeInterval(StartTiming(), EndTiming()),
        Exists(
            And(
                Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                Or(
                    is_connected(l_from, mid_location),
                    is_connected(mid_location, l_from),
                ),
                Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to)),
            ),
            mid_location,
        ),
    )
    move.add_condition(
        StartTiming(),
        Exists(
            And(
                Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                Or(
                    is_connected(l_from, mid_location),
                    is_connected(mid_location, l_from),
                ),
                Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to)),
            ),
            mid_location,
        ),
    )
    move.add_effect(StartTiming(1), is_at(l_from), False)
    move.add_effect(EndTiming(5), is_at(l_to), True)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    l3 = Object("l3", Location)
    l4 = Object("l4", Location)
    l5 = Object("l5", Location)
    problem = Problem("timed_connected_locations")
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.add_goal(is_at(l5))
    plan = up.plans.TimeTriggeredPlan(
        [
            (
                Fraction(0, 1),
                up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l3))),
                Fraction(6, 1),
            ),
            (
                Fraction(6, 1),
                up.plans.ActionInstance(move, (ObjectExp(l3), ObjectExp(l5))),
                Fraction(6, 1),
            ),
        ]
    )
    timed_connected_locations = Example(problem=problem, plan=plan)
    problems["timed_connected_locations"] = timed_connected_locations

    # hierarchical blocks world
    Entity = UserType("Entity", None)  # None can be avoided
    Location = UserType("Location", Entity)
    Unmovable = UserType("Unmovable", Location)
    TableSpace = UserType("TableSpace", Unmovable)
    Movable = UserType("Movable", Location)
    Block = UserType("Block", Movable)
    clear = Fluent("clear", BoolType(), space=Location)
    on = Fluent("on", BoolType(), object=Movable, space=Location)

    move = InstantaneousAction("move", item=Movable, l_from=Location, l_to=Location)
    item = move.parameter("item")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem("hierarchical_blocks_world")
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object("ts_1", TableSpace)
    ts_2 = Object("ts_2", TableSpace)
    ts_3 = Object("ts_3", TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object("block_1", Block)
    block_2 = Object("block_2", Block)
    block_3 = Object("block_3", Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    problem.add_goal(on(block_2, block_1))
    problem.add_goal(on(block_3, block_2))

    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))
            ),
            up.plans.ActionInstance(
                move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2))
            ),
        ]
    )
    hierarchical_blocks_world = Example(problem=problem, plan=plan)
    problems["hierarchical_blocks_world"] = hierarchical_blocks_world

    # robot with action duration expressed using static fluents
    problem = Problem("robot_with_durative_action")

    Location = UserType("Location")
    Robot = UserType("Robot")

    is_at = Fluent("is_at", BoolType(), position=Location, robot=Robot)
    is_connected = Fluent("is_connected", BoolType(), l_from=Location, l_to=Location)
    distance = Fluent("distance", RealType(), l_from=Location, l_to=Location)
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_fluent(distance, default_initial_value=1)

    move = DurativeAction("move", r=Robot, l_from=Location, l_to=Location)
    r = move.parameter("r")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.set_fixed_duration((distance(l_from, l_to)))
    move.add_condition(StartTiming(), is_connected(l_from, l_to))
    move.add_condition(StartTiming(), is_at(l_from, r))
    move.add_condition(StartTiming(), Not(is_at(l_to, r)))
    move.add_effect(StartTiming(), is_at(l_from, r), False)
    move.add_effect(EndTiming(), is_at(l_to, r), True)
    problem.add_action(move)

    r1 = Object("r1", Robot)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    l3 = Object("l3", Location)
    l4 = Object("l4", Location)
    l5 = Object("l5", Location)
    problem.add_objects([r1, l1, l2, l3, l4, l5])

    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.set_initial_value(distance(l1, l2), 10)
    problem.set_initial_value(distance(l2, l3), 10)
    problem.set_initial_value(distance(l3, l4), 10)
    problem.set_initial_value(distance(l4, l5), 10)

    problem.add_goal(is_at(l5, r1))

    plan = up.plans.TimeTriggeredPlan(
        [
            (
                Fraction(0, 1),
                up.plans.ActionInstance(
                    move, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l2))
                ),
                Fraction(10, 1),
            ),
            (
                Fraction(10, 1),
                up.plans.ActionInstance(
                    move, (ObjectExp(r1), ObjectExp(l2), ObjectExp(l3))
                ),
                Fraction(10, 1),
            ),
            (
                Fraction(10, 1),
                up.plans.ActionInstance(
                    move, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l4))
                ),
                Fraction(10, 1),
            ),
            (
                Fraction(10, 1),
                up.plans.ActionInstance(
                    move, (ObjectExp(r1), ObjectExp(l4), ObjectExp(l5))
                ),
                Fraction(10, 1),
            ),
        ]
    )
    robot_static_fluents_duration = Example(problem=problem, plan=plan)
    problems["robot_with_static_fluents_duration"] = robot_static_fluents_duration

    return problems



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up


class PlanValidatorMixin:
    @staticmethod
    def is_plan_validator() -> bool:
        return True

    @staticmethod
    def supports_plan(plan_kind: "up.plans.PlanKind") -> bool:
        raise NotImplementedError

    def validate(
        self, problem: "up.model.AbstractProblem", plan: "up.plans.Plan"
    ) -> "up.engines.results.ValidationResult":
        assert isinstance(self, up.engines.engine.Engine)
        if not self.supports(problem.kind):
            raise up.exceptions.UPUsageError(
                f"{self.name} cannot validate this kind of problem!"
            )
        if not self.supports_plan(plan.kind):
            raise up.exceptions.UPUsageError(
                f"{self.name} cannot validate this kind of plan!"
            )
        return self._validate(problem, plan)

    def _validate(
        self, problem: "up.model.AbstractProblem", plan: "up.plans.Plan"
    ) -> "up.engines.results.ValidationResult":
        raise NotImplementedError



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module implements the grounder that uses tarski."""


import shutil
from typing import Optional, Tuple, Dict, List
import tarski
import unified_planning as up
import unified_planning.interop
from unified_planning.interop.from_tarski import convert_tarski_formula
from unified_planning.model import Action, FNode, Problem, ProblemKind
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.engine import Engine, Credits
from unified_planning.engines.results import CompilerResult
from unified_planning.engines.compilers.grounder import Grounder
from tarski.grounding import LPGroundingStrategy


gringo = shutil.which("gringo")
if gringo is None:
    raise ImportError("Tarski grounder needs gringo installed")

credits = Credits(
    "Tarski grounder",
    "Artificial Intelligence and Machine Learning Group - Universitat Pompeu Fabra",
    "info-ai@upf.edu",
    "https://github.com/aig-upf/tarski",
    "Apache 2.0",
    "Tarski grounder, more information available on the given website.",
    "Tarski grounder, more information available on the given website.",
)


class TarskiGrounder(Engine, CompilerMixin):
    """Implements the gounder that uses tarski."""

    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            raise

    @property
    def name(self) -> str:
        return "tarski_grounder"

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind: "unified_planning.model.ProblemKind") -> bool:
        return problem_kind <= TarskiGrounder.supported_kind()

    @staticmethod
    def supports_compilation(compilation_kind: CompilationKind) -> bool:
        return compilation_kind == CompilationKind.GROUNDING

    def _compile(
        self, problem: "up.model.AbstractProblem", compilation_kind: "CompilationKind"
    ) -> CompilerResult:
        assert isinstance(problem, Problem)

        tarski_problem = up.interop.convert_problem_to_tarski(problem)
        actions = None
        try:
            lpgs = LPGroundingStrategy(tarski_problem)
            actions = lpgs.ground_actions()
        except tarski.grounding.errors.ReachabilityLPUnsolvable:
            raise up.exceptions.UPUsageError(
                "tarski grounder can not find a solvable grounding."
            )
        grounded_actions_map: Dict[Action, List[Tuple[FNode, ...]]] = {}
        fluents = {fluent.name: fluent for fluent in problem.fluents}
        objects = {object.name: object for object in problem.all_objects}
        types: Dict[str, Optional["up.model.Type"]] = {}
        if not problem.has_type("object"):
            types[
                "object"
            ] = None  # we set object as None, so when it is the father of a type in tarski, in UP it will be None.
        for action_name, list_of_tuple_of_parameters in actions.items():
            action = problem.action(action_name)
            parameters = {parameter.name: parameter for parameter in action.parameters}
            grounded_actions_map[action] = []
            for tuple_of_parameters in list_of_tuple_of_parameters:
                temp_list_of_converted_parameters = []
                for p in tuple_of_parameters:
                    if isinstance(p, str):
                        temp_list_of_converted_parameters.append(
                            problem.env.expression_manager.ObjectExp(problem.object(p))
                        )
                    else:
                        temp_list_of_converted_parameters.append(
                            convert_tarski_formula(
                                problem.env, fluents, objects, parameters, types, p
                            )
                        )
                grounded_actions_map[action].append(
                    tuple(temp_list_of_converted_parameters)
                )
        up_grounder = Grounder(grounding_actions_map=grounded_actions_map)
        up_res = up_grounder.compile(problem, compilation_kind)
        return CompilerResult(
            up_res.problem, up_res.map_back_action_instance, self.name
        )

    @staticmethod
    def get_credits(**kwargs) -> Optional[Credits]:
        return credits

    def destroy(self):
        pass



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from unified_planning.environment import Environment
from unified_planning.model.fnode import FNode
from enum import Enum, auto
from fractions import Fraction
from typing import Union, Optional


class TimepointKind(Enum):
    GLOBAL_START = auto()
    GLOBAL_END = auto()
    START = auto()
    END = auto()


class Timepoint:
    def __init__(self, kind: TimepointKind, container: Optional[str] = None):
        """Creates a new timepoint.

        It is typically used to refer to:
         - the start/end of the containing action or method, or
         - to the start/end of a subtasks in a method

        Parameters
        ----------
        kind: TimepointKind
          Kind of the timepoint.
        container: Optional[str]
          Identifier of the container in which the timepoint is defined.
          If not set, then a start/end timepoint refers to the enclosing action or method.
        """
        assert container is None or isinstance(container, str)
        self._kind = kind
        self._container = container

    def __repr__(self):
        if (
            self._kind == TimepointKind.GLOBAL_START
            or self._kind == TimepointKind.START
        ):
            qualifier = "start"
        else:
            qualifier = "end"
        if self._container is None:
            return qualifier
        else:
            return f"{qualifier}({self._container})"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Timepoint):
            return self._kind == oth._kind and self._container == oth._container
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._kind) + hash(self._container)

    @property
    def kind(self) -> TimepointKind:
        return self._kind

    @property
    def container(self):
        """Returns the container in which this timepoint is defined or None if it refers to the enclosing action/method."""
        return self._container


class Timing:
    def __init__(self, delay: Union[int, Fraction], timepoint: Timepoint):
        self._timepoint = timepoint
        self._delay = delay

    def __repr__(self):
        if self._delay == 0:
            return f"{self._timepoint}"
        else:
            return f"{self._timepoint} + {self._delay}"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Timing):
            return self._delay == oth._delay and self._timepoint == oth._timepoint
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._delay) ^ hash(self._timepoint)

    @property
    def delay(self) -> Union[int, Fraction]:
        return self._delay

    @property
    def timepoint(self) -> Timepoint:
        return self._timepoint

    def is_global(self) -> bool:
        return (
            self._timepoint.kind == TimepointKind.GLOBAL_START
            or self._timepoint.kind == TimepointKind.GLOBAL_END
        )

    def is_from_start(self) -> bool:
        return (
            self._timepoint.kind == TimepointKind.START
            or self._timepoint.kind == TimepointKind.GLOBAL_START
        )

    def is_from_end(self) -> bool:
        return not self.is_from_start()


def StartTiming(
    delay: Union[int, Fraction] = 0, container: Optional[str] = None
) -> Timing:
    """Represents the start timing of an action.
    Created with a delay > 0 represents "delay" time
    after the start of an action.

    For example, action starts at time 5:
    StartTiming() = 5
    StartTiming(3) = 5+3 = 8"""

    return Timing(delay, Timepoint(TimepointKind.START, container=container))


def EndTiming(
    delay: Union[int, Fraction] = 0, container: Optional[str] = None
) -> Timing:
    """Represents the end timing of an action.
    Created with a delay > 0 represents "delay" time
    before the end of an action.

    For example, action ends at time 10:
    EndTiming() = 10
    EndTiming(1.5) = 10-Fraction(3, 2) = Fraction(17, 2) = 8.5"""

    return Timing(delay, Timepoint(TimepointKind.END, container=container))


def GlobalStartTiming(delay: Union[int, Fraction] = 0):
    """Represents the absolute timing.
    Created with a delay > 0 represents "delay" time
    after the start of the execution."""

    return Timing(delay, Timepoint(TimepointKind.GLOBAL_START))


def GlobalEndTiming(delay: Union[int, Fraction] = 0):
    """Represents the end timing of an execution.
    Created with a delay > 0 represents "delay" time
    before the end of the execution."""

    return Timing(delay, Timepoint(TimepointKind.GLOBAL_END))


class Interval:
    def __init__(
        self,
        lower: FNode,
        upper: FNode,
        is_left_open: bool = False,
        is_right_open: bool = False,
    ):
        self._lower = lower
        self._upper = upper
        self._is_left_open = is_left_open
        self._is_right_open = is_right_open
        assert (
            lower.environment == upper.environment
        ), "Interval s boundaries expression can not have different environments"

    def __repr__(self) -> str:
        if self.is_left_open():
            left_bound = "("
        else:
            left_bound = "["
        if self.is_right_open():
            right_bound = ")"
        else:
            right_bound = "]"
        return f"{left_bound}{str(self.lower)}, {str(self.upper)}{right_bound}"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Interval):
            return (
                self._lower == oth._lower
                and self._upper == oth._upper
                and self._is_left_open == oth._is_left_open
                and self._is_right_open == oth._is_right_open
            )
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._lower) + hash(self._upper)
        if self._is_left_open:
            res ^= hash("is_left_open")
        if self._is_right_open:
            res ^= hash("is_right_open")
        return res

    @property
    def lower(self) -> FNode:
        return self._lower

    @property
    def upper(self) -> FNode:
        return self._upper

    @property
    def environment(self) -> "Environment":
        return self._lower.environment

    def is_left_open(self) -> bool:
        return self._is_left_open

    def is_right_open(self) -> bool:
        return self._is_right_open


class Duration:
    pass


class DurationInterval(Duration, Interval):
    def __init__(
        self,
        lower: FNode,
        upper: FNode,
        is_left_open: bool = False,
        is_right_open: bool = False,
    ):
        Duration.__init__(self)
        Interval.__init__(self, lower, upper, is_left_open, is_right_open)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, DurationInterval):
            return (
                self._lower == oth._lower
                and self._upper == oth._upper
                and self._is_left_open == oth._is_left_open
                and self._is_right_open == oth._is_right_open
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash((self._lower, self.upper, self._is_left_open, self._is_right_open))


def ClosedDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    """Represents the (closed) interval duration constraint:
    [lower, upper]
    """
    return DurationInterval(lower, upper)


def FixedDuration(size: FNode) -> DurationInterval:
    """Represents a fixed duration constraint"""
    return DurationInterval(size, size)


def OpenDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    """Represents the (open) interval duration constraint:
    (lower, upper)
    """
    return DurationInterval(lower, upper, True, True)


def LeftOpenDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    """Represents the (left open, right closed) interval duration constraint:
    (lower, upper]
    """
    return DurationInterval(lower, upper, True, False)


def RightOpenDurationInterval(lower: FNode, upper: FNode) -> DurationInterval:
    """Represents the (left closed, right open) interval duration constraint:
    [lower, upper)
    """
    return DurationInterval(lower, upper, False, True)


class TimeInterval:
    def __init__(
        self,
        lower: Timing,
        upper: Timing,
        is_left_open: bool = False,
        is_right_open: bool = False,
    ):
        self._lower = lower
        self._upper = upper
        self._is_left_open = is_left_open
        self._is_right_open = is_right_open

    def __repr__(self) -> str:
        if self.is_left_open():
            left_bound = "("
        else:
            left_bound = "["
        if self.is_right_open():
            right_bound = ")"
        else:
            right_bound = "]"
        if self.lower == self.upper:
            return f"{left_bound}{str(self.lower)}{right_bound}"
        else:
            return f"{left_bound}{str(self.lower)}, {str(self.upper)}{right_bound}"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, TimeInterval):
            return (
                self._lower == oth._lower
                and self._upper == oth._upper
                and self._is_left_open == oth._is_left_open
                and self._is_right_open == oth._is_right_open
            )
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._lower) + hash(self._upper)
        if self._is_left_open:
            res ^= hash("is_left_open")
        if self._is_right_open:
            res ^= hash("is_right_open")
        return res

    @property
    def lower(self) -> Timing:
        return self._lower

    @property
    def upper(self) -> Timing:
        return self._upper

    def is_left_open(self) -> bool:
        return self._is_left_open

    def is_right_open(self) -> bool:
        return self._is_right_open


def TimePointInterval(tp: Timing) -> TimeInterval:
    """Represents the (point) temporal interval: [tp, tp]"""
    return TimeInterval(tp, tp)


def ClosedTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    """Represents the (closed) temporal interval: [lower, upper]"""
    return TimeInterval(lower, upper)


def OpenTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    """Represents the (open) temporal interval: (lower, upper)"""
    return TimeInterval(lower, upper, True, True)


def LeftOpenTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    """Represents the (left open, right closed) temporal interval: (lower, upper]"""
    return TimeInterval(lower, upper, True, False)


def RightOpenTimeInterval(lower: Timing, upper: Timing) -> TimeInterval:
    """Represents the (left closed, right open) temporal interval: [lower, upper)"""
    return TimeInterval(lower, upper, False, True)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import unified_planning.grpc.generated.unified_planning_pb2 as unified__planning__pb2


class UnifiedPlanningStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.planOneShot = channel.unary_stream(
                '/UnifiedPlanning/planOneShot',
                request_serializer=unified__planning__pb2.PlanRequest.SerializeToString,
                response_deserializer=unified__planning__pb2.PlanGenerationResult.FromString,
                )
        self.validatePlan = channel.unary_unary(
                '/UnifiedPlanning/validatePlan',
                request_serializer=unified__planning__pb2.ValidationRequest.SerializeToString,
                response_deserializer=unified__planning__pb2.ValidationResult.FromString,
                )
        self.compile = channel.unary_unary(
                '/UnifiedPlanning/compile',
                request_serializer=unified__planning__pb2.Problem.SerializeToString,
                response_deserializer=unified__planning__pb2.CompilerResult.FromString,
                )


class UnifiedPlanningServicer(object):
    """Missing associated documentation comment in .proto file."""

    def planOneShot(self, request, context):
        """A plan request to the engine.
        The engine replies with a stream of N `Answer` messages where:
        - the first (N-1) message are of type `IntermediateReport`
        - the last message is of type `FinalReport`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def validatePlan(self, request, context):
        """A validation request to the engine.
        The engine replies with the ValidationResult
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def compile(self, request, context):
        """A compiler request to the engine.
        The engine replies with the CompilerResult
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UnifiedPlanningServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'planOneShot': grpc.unary_stream_rpc_method_handler(
                    servicer.planOneShot,
                    request_deserializer=unified__planning__pb2.PlanRequest.FromString,
                    response_serializer=unified__planning__pb2.PlanGenerationResult.SerializeToString,
            ),
            'validatePlan': grpc.unary_unary_rpc_method_handler(
                    servicer.validatePlan,
                    request_deserializer=unified__planning__pb2.ValidationRequest.FromString,
                    response_serializer=unified__planning__pb2.ValidationResult.SerializeToString,
            ),
            'compile': grpc.unary_unary_rpc_method_handler(
                    servicer.compile,
                    request_deserializer=unified__planning__pb2.Problem.FromString,
                    response_serializer=unified__planning__pb2.CompilerResult.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'UnifiedPlanning', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UnifiedPlanning(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def planOneShot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/UnifiedPlanning/planOneShot',
            unified__planning__pb2.PlanRequest.SerializeToString,
            unified__planning__pb2.PlanGenerationResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def validatePlan(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UnifiedPlanning/validatePlan',
            unified__planning__pb2.ValidationRequest.SerializeToString,
            unified__planning__pb2.ValidationResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def compile(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UnifiedPlanning/compile',
            unified__planning__pb2.Problem.SerializeToString,
            unified__planning__pb2.CompilerResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines an Object of a planning problem.
An Object is represented by a name and by its type.
"""

from typing import Optional
from unified_planning.environment import Environment, get_env
import unified_planning.model.types


class Object:
    """Represents an object."""

    def __init__(
        self,
        name: str,
        typename: "unified_planning.model.types.Type",
        env: Optional[Environment] = None,
    ):
        self._name = name
        self._typename = typename
        self._env = get_env(env)
        assert self._env.type_manager.has_type(
            typename
        ), "type of the object does not belong to the same environment of the object"

    def __repr__(self) -> str:
        return self.name

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Object):
            return (
                self._name == oth._name
                and self._typename == oth._typename
                and self._env == oth._env
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._name) + hash(self._typename)

    @property
    def name(self) -> str:
        """Returns the object name."""
        return self._name

    @property
    def type(self) -> "unified_planning.model.types.Type":
        """Returns the object type."""
        return self._typename

    @property
    def environment(self) -> "Environment":
        """Return the object environment"""
        return self._env

    #
    # Infix operators
    #

    def Equals(self, right):
        return self._env.expression_manager.Equals(self, right)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.test.examples import get_example_problems
from unified_planning.io import PythonWriter


class TestPythonWriter(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_all(self):
        for p in self.problems.values():
            original_problem = p.problem
            pw = PythonWriter(original_problem)
            _locals = {}
            exec(pw.write_problem_code(), {}, _locals)
            self.assertEqual(original_problem, _locals["problem"])
            self.assertEqual(hash(original_problem), hash(_locals["problem"]))

    def test_ad_hoc_1(self):
        xd = Fluent("x-$")
        xe = Fluent("x-&")
        nop = Fluent("")
        a = InstantaneousAction("3")
        a.add_precondition(Not(xd))
        a.add_effect(xd, True)
        a.add_precondition(Not(xe))
        a.add_effect(xe, True)
        a.add_precondition(Not(nop))
        a.add_effect(nop, True)
        problem = Problem("basic")
        problem.add_fluent(xd)
        problem.add_fluent(xe)
        problem.add_fluent(nop)
        problem.add_action(a)
        problem.set_initial_value(xd, False)
        problem.set_initial_value(xe, False)
        problem.set_initial_value(nop, False)
        problem.add_goal(xd)
        problem.add_goal(xe)
        problem.add_goal(nop)
        pw = PythonWriter(problem)
        _locals = {}
        exec(pw.write_problem_code(), {}, _locals)
        self.assertEqual(problem, _locals["problem"])

    def test_ad_hoc_2(self):
        Location = UserType("Location")
        robot_at = Fluent("robot_at", BoolType(), OrderedDict([("if", Location)]))
        battery_charge = Fluent("battery_charge", RealType(0, 100))
        move = InstantaneousAction(
            "move", OrderedDict([("from", Location), ("to", Location)])
        )
        l_from = move.parameter("from")
        to = move.parameter("to")
        move.add_precondition(GE(battery_charge, 10))
        move.add_precondition(Not(Equals(l_from, to)))
        move.add_precondition(robot_at(l_from))
        move.add_precondition(Not(robot_at(to)))
        move.add_effect(robot_at(l_from), False)
        move.add_effect(robot_at(to), True)
        move.add_effect(battery_charge, Minus(battery_charge, 10))
        l1 = Object("l1", Location)
        l2 = Object("l2", Location)
        problem = Problem("robot")
        problem.add_fluent(robot_at)
        problem.add_fluent(battery_charge)
        problem.add_action(move)
        problem.add_object(l1)
        problem.add_object(l2)
        problem.set_initial_value(robot_at(l1), True)
        problem.set_initial_value(robot_at(l2), False)
        problem.set_initial_value(battery_charge, 100)
        problem.add_goal(robot_at(l2))
        pw = PythonWriter(problem)
        _locals = {}
        exec(pw.write_problem_code(), {}, _locals)
        self.assertEqual(problem, _locals["problem"])



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the quantifiers remover class."""


import unified_planning.model.walkers as walkers
from unified_planning.model.walkers.identitydag import IdentityDagWalker
from unified_planning.model.fnode import FNode
from unified_planning.model.operators import OperatorKind
from unified_planning.model.object import Object
from unified_planning.model.mixins import ObjectsSetMixin
from unified_planning.model.expression import Expression
from typing import List, Dict
from itertools import product


class ExpressionQuantifiersRemover(IdentityDagWalker):
    def __init__(self, env):
        self._env = env
        IdentityDagWalker.__init__(self, self._env, True)
        self._substituter = walkers.substituter.Substituter(self._env)

    def remove_quantifiers(self, expression: FNode, objects_set: "ObjectsSetMixin"):
        self._objects_set = objects_set
        return self.walk(expression)

    def _help_walk_quantifiers(
        self, expression: FNode, args: List[FNode]
    ) -> List[FNode]:
        vars = expression.variables()
        type_list = [v.type for v in vars]
        possible_objects: List[List[Object]] = [
            list(self._objects_set.objects(t)) for t in type_list
        ]
        # product of n iterables returns a generator of tuples where
        # every tuple has n elements and the tuples make every possible
        # combination of 1 item for each iterable. For example:
        # product([1,2], [3,4], [5,6], [7]) =
        # (1,3,5,7) (1,3,6,7) (1,4,5,7) (1,4,6,7) (2,3,5,7) (2,3,6,7) (2,4,5,7) (2,4,6,7)
        subs_results = []
        for o in product(*possible_objects):
            subs: Dict[Expression, Expression] = dict(zip(vars, list(o)))
            subs_results.append(self._substituter.substitute(args[0], subs))
        return subs_results

    @walkers.handles(OperatorKind.EXISTS)
    def walk_exists(self, expression: FNode, args: List[FNode], **kwargs) -> FNode:
        subs_results = self._help_walk_quantifiers(expression, args)
        return self._env.expression_manager.Or(subs_results)

    @walkers.handles(OperatorKind.FORALL)
    def walk_forall(self, expression: FNode, args: List[FNode], **kwargs) -> FNode:
        subs_results = self._help_walk_quantifiers(expression, args)
        return self._env.expression_manager.And(subs_results)



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
from fractions import Fraction
from typing import List, Union, Dict

import unified_planning as up
from unified_planning.model.htn.method import Method
from unified_planning.model.htn.task import Task
from unified_planning.model.htn.task_network import TaskNetwork


class HierarchicalProblem(up.model.problem.Problem):
    def __init__(
        self,
        name: str = None,
        env: "up.environment.Environment" = None,
        *,
        initial_defaults: Dict[
            "up.model.types.Type",
            Union[
                "up.model.fnode.FNode",
                "up.model.object.Object",
                bool,
                int,
                float,
                Fraction,
            ],
        ] = {},
    ):
        super().__init__(name=name, env=env, initial_defaults=initial_defaults)
        self._abstract_tasks: OrderedDict[str, Task] = OrderedDict()
        self._methods: OrderedDict[str, Method] = OrderedDict()
        self._initial_task_network = TaskNetwork()

    def __repr__(self):
        s = [super().__repr__()]
        s.append("abstract tasks = [\n")
        for t in self._abstract_tasks.values():
            s.append(f"  {t}\n")
        s.append("]\n\n")
        s.append("methods = [")
        for m in self._methods.values():
            s.append(("\n" + str(m)).replace("\n", "\n  "))
        s.append("\n]\n\n")
        s.append(str(self._initial_task_network))
        return "".join(s)

    def __eq__(self, oth: object) -> bool:
        if not super().__eq__(oth):
            return False
        if not isinstance(oth, HierarchicalProblem):
            return False
        return (
            self._initial_task_network == oth._initial_task_network
            and self._methods == oth._methods
            and self._abstract_tasks == oth._abstract_tasks
        )

    def __hash__(self):
        res = super().__hash__()
        res += sum(map(hash, self._abstract_tasks.values()))
        res += sum(map(hash, self._methods.values()))
        res += hash(self._initial_task_network)
        return res

    def clone(self):
        new_p = HierarchicalProblem(self._name, self._env)
        new_p._fluents = self._fluents[:]
        new_p._actions = [a.clone() for a in self._actions]
        new_p._user_types = self._user_types[:]
        new_p._user_types_hierarchy = self._user_types_hierarchy.copy()
        new_p._objects = self._objects[:]
        new_p._initial_value = self._initial_value.copy()
        new_p._timed_effects = {
            t: [e.clone() for e in el] for t, el in self._timed_effects.items()
        }
        new_p._timed_goals = {i: [g for g in gl] for i, gl in self._timed_goals.items()}
        new_p._goals = self._goals[:]
        new_p._metrics = []
        for m in self._metrics:
            if isinstance(m, up.model.metrics.MinimizeActionCosts):
                costs = {new_p.action(a.name): c for a, c in m.costs.items()}
                new_p._metrics.append(up.model.metrics.MinimizeActionCosts(costs))
            else:
                new_p._metrics.append(m)
        new_p._initial_defaults = self._initial_defaults.copy()
        new_p._fluents_defaults = self._fluents_defaults.copy()
        new_p._initial_task_network = self._initial_task_network.clone()
        new_p._methods = self._methods.copy()
        new_p._abstract_tasks = self._abstract_tasks.copy()
        return new_p

    @property
    def kind(self) -> "up.model.problem_kind.ProblemKind":
        """Returns the problem kind of this planning problem.

        IMPORTANT NOTE: this property does a lot of computation, so it should be called as
        minimum time as possible."""
        self._kind = super().kind
        self._kind.set_problem_class("HIERARCHICAL")
        return self._kind

    @property
    def tasks(self) -> List[Task]:
        return list(self._abstract_tasks.values())

    def get_task(self, task_name: str) -> Task:
        return self._abstract_tasks[task_name]

    def has_task(self, task_name: str):
        return task_name in self._abstract_tasks

    def add_task(self, task: Union[Task, str], **kwargs: "up.model.types.Type") -> Task:
        if isinstance(task, str):
            task = Task(task, _parameters=OrderedDict(**kwargs))
        else:
            assert len(kwargs) == 0
        assert (
            task.name not in self._abstract_tasks
        ), f"A task with name '{task.name}' already exists."
        self._abstract_tasks[task.name] = task
        return task

    @property
    def methods(self) -> List[Method]:
        return list(self._methods.values())

    def method(self, method_name) -> Method:
        return self._methods[method_name]

    def add_method(self, method: Method):
        assert (
            method.achieved_task is not None
        ), f"No achieved task was specified for this method."
        assert (
            method.name not in self._methods
        ), f"A method with name '{method.name}' already exists."
        assert (
            method.achieved_task.task.name in self._abstract_tasks
        ), f"Method is associated to an unregistered task '{method.achieved_task.task.name}'"
        self._methods[method.name] = method

    @property
    def task_network(self):
        return self._initial_task_network
